"""
Fluorescence Intensity calculations:
   Y. Choi

calculates fluorescence intensities for multiple elements at a fixed incident x-ray energy.
measured fluorescence depends on net fluorescence yield and transmission.
included are:
    (with incident energy dependence): atomic crosssection, fluorescence yield for each edge,
                                     transition probability for each emission line,
    (with emitted fluorescence energy dependence): transmission to vortex detector, detector efficiency.
global variables: re, Barn, Nav, pre_edge_margin, fluo_emit_min, det_res
AtNum2f1f2Xect: needs datafile readf1f2a.py which is from chantler table
cal_NetYield2: for XRF calculations.  major and minor lines are calculated
                output as 'Elemental_Sensitivity.txt'

sim_spectra: keeps individual emission lines with significant intensity without weight-averaging
             each element can have a different relative concentration.
             makes strongest emission to 100 and scales the rest accordingly.
             output as 'simSpectrum.txt'
             runs sim_GaussPeaks() to emulate measure intensity as gaussian peak
             output as 'simSpectrum.plot'
             runs cal_NetYield2()
5/24/2010:  self-absorption effect included, 4 options (top, bottom, all, surface)
            Each option makes different elemental distribution,
            but substrate elements are always distributed evenly.
            get_ChemName(), nominal_density() are in readf1f2a.py
5/26/2010:   incident x-ray attenuation effect added, and incident angle effect is also added.
            refraction effect added to the angle, below critical angle, th=1e-10.
6/4/2010:    SampleMatrix2 has two layers and each layer can have different composition
            Class ElemFY is for fluorescence elements with AtSym, Conc, tag attribues,
            inputs in SimSpect, Cal_NetFyield2 changed.
            Concentrations are normalized to the concentration of the substrate1 molecule.
            Ex) with CaCO3[2.71g/cc]/Fe2O3[5.26g/cc] substrate, Ca concentration is 1.0 meaning 1.63e22 Ca/cc
            and Fe concentration is 2.43 = 2*1.98e22/1.63e22 where 1.98e22 is number density (/cc) of Fe2O3
6/8/2010:    now using python2.6 instead of 2.5
4/14/2011:  fluo_det.py: detection dependent part in fluorescence yield.
            Original fluo.py was modified to fluo_new.py.
            readf1f2a.py is used instead of original readf1f2.py
            fluo_new.py is being split into two.
            fluo_elem.py is for elemtnal dependence: fluorescenc yield, crosssection.
            fluo_det.py is for detection dependnece: detector efficiency, attenuation, sample
            * for attenuation, total cross-section([4]) is used instead of photoelectric ([2]).
            ** for fluorescence yield, photoelectric is used.
"""


import math
import numpy
import sys
from larch.utils.physical_constants import AVOGADRO, BARN
from larch.xray import xray_delta_beta, chemparse

pre_edge_margin=150.    # FY calculated from 150 eV below the absorption edge.
fluo_emit_min=500.      # minimum energy for emitted fluorescence.  ignore fluorescence emissions below 500eV
det_res=100.            # detector resoltuion in eV, used in sim_GaussPeaks

'''
----------------------------------------------------------------------------------------------------------
class: Material, ElemFY, SampleMatrix2
----------------------------------------------------------------------------------------------------------
'''
class Material:
    def __init__(self, composition, density, thickness=1):
        self.composition=composition        # ex) SiO2
        self.density=density              # in g/cm^3
        self.thickness=thickness            # in cm
        self.la=1.0                         #absoprtion length in cm
        self.trans=0.5                      # transmission.
        self.absrp=0.5                 # absorption
        self.delta=0.1                  # index of refraction, real part
        self.beta=0.1                   # index of refraction, imaginary part
        # MN replace
        elements = chemparse(composition)
        AtomList = elements.keys()
        AtomIndex = elements.values()
        AtomWeight=0
        for (ii, atom) in enumerate(AtomList):
            # MN replace...
            AtWt= atomic_mass(atom)
            index=AtomIndex[ii]
            AtomWeight = AtomWeight + index*AtWt
        NumberDensity=density*AVOGARDRO/AtomWeight
        self.NumDen=NumberDensity       # number of molecules per cm^3
        self.AtWt=AtomWeight            # weight per mole
    def getLa(self, energy, NumLayer=1.0):    # get absorption legnth for incident x-ray, NumLayer for multiple layers
        # MN replace...
        temp= xray_delta_beta(self.composition, self.density, energy)
        # returns delta, beta, la_photoE, Nat, la_total
        self.delta=temp[0];     self.beta=temp[1]
        self.la=temp[2]  # temp[4] (total) instead of temp[2] (photoelectric)
        if NumLayer<0:  NumLayer=0     # Number of layers cannot be less than zero
        self.trans=math.exp(-self.thickness*NumLayer/self.la)  #la attenuation length cm, NumLayer for multiple layers
        self.absrp=1-self.trans


class ElemFY:  # fluorescing element
    def __init__(self, AtomicSymbol='Fe', Concentration=10e-6, tag='dilute'):
        temp=AtomicSymbol[0].upper()  # capitalize the first letter
        self.AtSym=temp+AtomicSymbol[1:]
        self.Conc=Concentration
        self.tag=tag  # dilute, substrate1, substrate2


class SampleMatrix2:  # sample matrix for self-absorption correction, 6/3: two layers with different compositions
    def __init__(self, composition1='Si', density1=2.33, thickness1=0.1, \
                 composition2='Si', density2=2.33, thickness2=0.1,
                 angle0=45.0, option='surface'):
        self.composition1 = composition1     # ex) Fe2O3
        # MN replace:
        out= chemparse(composition1)         # output from get_ChemName
        self.ElemList1 = out.keys()          # list of elments in matrix: 'Fe', 'O' for Fe2O3
        self.ElemInd1 = out.values()         # list of index: 2, 3 for Fe2O3
        # self.ElemFrt1 = out[2]             # list of fraction: 0.4, 0.6 for Fe2O3
        self.density1 = density1             # in g/cm^3
        self.thickness1 = thickness1         # top layer thickness in cm
        self.composition2 = composition2
        # MN replace with chemparse()
        out=chemparse(composition2)
        self.ElemList2 = out.keys()              # for the bottom substrate
        self.ElemInd2 = out.values()
        # self.ElemFrt2 = out[2]
        self.density2 = density2
        self.thickness2 = thickness2         # bottom layer thickness in cm
        self.angle = angle0*math.pi/180.     # in radian, incident beam angle, surface normal =pi/2
        self.option = option                 # option for fluorescing element location, surface/top/bottom
        self.la1 = 1.0                       # absoprtion length in cm
        self.delta1 = 0.1                    # index of refraction, real part correction
        self.beta1 = 0.1                     # index of refraction, imaginary part
        self.Nat1 = 1.0                      # atomic number density of the Fe2O3, 1e22 Fe2O3 atoms /cm^3
        self.la2 = 1.0
        self.delta2 = 0.1
        self.beta2 = 0.1
        self.Nat2 = 1.0                     # atomic number density of the first element, 1e22 Fe2O3 atoms/cm^3
        self.scale = 1.0                    # weighted average over the depth range: sum of (trans*factors)
        self.scale1 = 1.0                   # sum of (trans*1.0) this is for substrate element in top layer
        self.scale2 = 1.0                   # sum of (trans*thickness2/thickness1) for substrate element in bottom layer
        self.ElemListFY =[]                 # fluorescing element
        self.ElemFrtFY =[]                  # fraction for fluorescing element
        self.Nat1=1                  # atomic number density in atoms/cc, for example # of Fe2O3/cm^3 for Fe2O3 substrate
        self.Nat2=1                  # atomic number density in atoms/cc
        substrate1_material = Material(composition1, density1)
        AtNumDen1 = substrate1_material.NumDen      #  substrate1
        substrate2_material = Material(composition2, density2)
        AtNumDen2 = substrate2_material.NumDen      #  substrate2, fixed 8/12/10
        self.Nat1=AtNumDen1
        self.Nat2=AtNumDen2
        self.txt=''
        text1='substrate1:%6.3e %s/cm^3' % (AtNumDen1, composition1)
        #print(text1)
        text2='substrate2:%6.3e %s/cm^3' % (AtNumDen2, composition2)
        self.txt=text1+'  '+text2
        print(self.txt)
        # atom.conc is normalized to the number density of substrate1 molecule
        for (ii, item) in enumerate(self.ElemList1):
            # MN replace:
            if atomic_number(item)>=12:       # ignore elements below Mg
                #atom=ElemFY(item, self.ElemFrt1[ii], 'substrate1')
                atom=ElemFY(item, self.ElemInd1[ii], 'substrate1')
                # eg. for Fe2O3, Fe concentration is 2 (=2 x Fe2O3 number density)
                self.ElemListFY.append(atom)
        for (ii, item) in enumerate(self.ElemList2):
            # MN replace:
            if atomic_number(item)>=12:       # ignore elements below Mg
                #atom=ElemFY(item, self.ElemFrt2[ii], 'substrate2')
                atom=ElemFY(item, self.ElemInd2[ii]*AtNumDen2/AtNumDen1, 'substrate2')
                self.ElemListFY.append(atom)
        numLayer=100                        # each layer is sliced into 100 sublayers.
        self.depths=[]                      # depth values for calculation
        for ii in range(numLayer):
            step=1.0/numLayer
            self.depths.append(step*ii*self.thickness1)
        for ii in range(numLayer):
            step=1.0/numLayer
            self.depths.append(step*ii*self.thickness2+self.thickness1)
        self.factors=[]                     # 1 or pre if element present at each depth
        pre=self.thickness2/self.thickness1 # prefactor, if two layers have different thickness
        if self.option=='surface':          # fluorescecing atoms present in only on the surface
            for ii in range(len(self.depths)):
                if ii==0:
                    self.factors.append(1.0)
                else:
                    self.factors.append(0.0)
        if self.option=='all':              # fluorescecing atoms present throughout
            for ii in range(len(self.depths)):
                if self.depths[ii]<self.thickness1:
                    self.factors.append(1.0)
                else:
                    self.factors.append(pre)
        if self.option=='top':              # fluorescecing atoms present only in the top layer
            for ii in range(len(self.depths)):
                if self.depths[ii]<self.thickness1:
                    self.factors.append(1.0)
                else:
                    self.factors.append(0.0)
        if self.option=='bottom':           # fluorescecing atoms present only in the bottom layer
            for ii in range(len(self.depths)):
                if self.depths[ii]<self.thickness1:
                    self.factors.append(0.0)
                else:
                    self.factors.append(pre)
        # note: fluorescing substrate atoms present throughout regardless of the option.
        self.trans=[]                       # transmission to surface for emitted fluorescence
        self.absrp=[]                       # absorption until surface for emitted fluorescence
        self.inten0=[]                      # incident x-ray intensity at each depth
        for ii in range(len(self.depths)):
            self.trans.append(0.5)
            self.absrp.append(0.5)
            self.inten0.append(0.5)
    def getPenetration(self, energy0):  # incident x-ray penetration(attenuation)
        # refraction at air/top layer interface
        # MN replace:
        temp=f1f2.get_delta(self.composition1, self.density1, energy0)  # energy0: incident x-ray energy
        delta1=temp[0];  beta1=temp[1]
        la1=temp[4]     # in cm, temp[4] instead of temp[2], using total instead of photoelectric
        self.la1=la1                                        # absorption length in microns at incident x-ray energy
        angle_critical1 = (2.0*delta1)**(0.5)               # in radian, critical angle for total external reflection
        if angle_critical1>=self.angle:                     # below critical angle, the corrected should be zero.
            angle_corrected1=1.0e-15                        # a smaller number instead of zero
        else:                                               # above critical angle
            angle_corrected1 = (self.angle**2.0 - angle_critical1**2.0)**(0.5)  # in radian
        # refraction at top/bottom layers interface
        # MN replace:
        temp=f1f2.get_delta(self.composition2, self.density2, energy0)  # energy0: incident x-ray energy
        delta2=temp[0];  beta2=temp[1];
        la2=temp[4]     # in cm, temp[4] instead of temp[2], using total instead of photoelectric
        self.la2=la2                                            # absorption length in cm at incident x-ray energy
        angle_corrected2 = ( 2.0-(1.0-delta1)/(1.0-delta2)*(2.0-angle_corrected1**2) )**0.5
        # using Snell's law, assume beta effect not ignificant, in radian
        for (ii, depth) in enumerate(self.depths):
            if self.depths[ii]<self.thickness1:                 # top layer
                beampath = depth/math.sin(angle_corrected1)     # in cm
                inten0 = math.exp(-beampath/la1)                # attenuated incident beam intensity at depths
            else:                                               # bottom layer
                beampath1 = self.thickness1/math.sin(angle_corrected1)
                beampath2 = (depth-self.thickness1)/math.sin(angle_corrected2)
                inten0 = math.exp(-beampath1/la1)*math.exp(-beampath2/la2)
            self.inten0[ii] = inten0                            # incident x-ray attenuation
    def getLa(self, energy, NumLayer=1.0):  # emitted fluorescence trasmission attenuation up to top surface
        transmitted=1.0
        # MN replace:
        temp=f1f2.get_delta(self.composition1, self.density1, energy)  # energy is for fluorescence
        self.delta1 = temp[0]
        self.beta1 = temp[1]
        self.la1 = temp[4]      # in cm, temp[4] instead of temp[2], using total instead of photoelectric
        # absorption length in cm at emitted fluorescence energy
        # temp[3]: atomic number density of the first element in atoms/cc
        # MN replace:
        temp=f1f2.get_delta(self.composition2, self.density2, energy)  # energy is for fluorescence
        self.delta2 = temp[0]
        self.beta2 = temp[1]
        self.la2 = temp[4]                                      # absorption length in cm at emitted fluorescence energy
        # in cm, temp[4] instead of temp[2], using total instead of photoelectric
        angle_exit = (math.pi/2.0 - self.angle)                 # becomes 90 at a small incident angle
        for (ii, depth) in enumerate(self.depths):
            if self.depths[ii]<self.thickness1:                 # top layer
                 transmitted=math.exp(-depth/math.sin(angle_exit)*NumLayer/self.la1)
            else:                                               # bottom layer
                transmitted2 = math.exp(-(depth-self.thickness1)/math.sin(angle_exit)*NumLayer/self.la2)
                transmitted1 = math.exp(-self.thickness1/math.sin(angle_exit)*NumLayer/self.la1)
                transmitted = transmitted2*transmitted1
        self.trans[ii] = transmitted
        self.absrp[ii] = 1.0 - transmitted
        scale = 0.0; scale1=0.0; scale2=0.0
        for (ii,trans) in enumerate(self.trans):
            scale = scale + trans*self.inten0[ii]*self.factors[ii]
            if self.depths[ii]<self.thickness1:
                scale1 = scale1 + trans*self.inten0[ii]*1.0
            else:   # if thickness2 is different from thickness1, weight differently
                scale2 = scale2 + trans*self.inten0[ii]*(self.thickness2/self.thickness1)
        # scale, scale1, scale2: emitted fluorescence transmission and depth profile
        self.scale = scale      # sum of (trans*factors) for nonsubstrate FY
        self.scale1 = scale1    # sum of (trans*factors) for substrate FY. factors=1, top layer
        self.scale2 = scale2    # sum of (trans*factors) for substrate FY. factors=pre, bttom layer


'''
----------------------------------------------------------------------------------------------------------
Detector efficiency, fluorescence attenuation
----------------------------------------------------------------------------------------------------------
'''
# WD30 used for XSW, sample-->He-->Kapton-->Collimator-->Detector
# WD60 used for XRM, sample-->Collimator-->Detector
# eV1 is fluorescence energy in eV
# xHe=1 means Helium gas from sample to WD60 collimator.
# xAl=1 means 1 layer of 1.5mil Al foil as attenuator
# xKapton=1 means 1 layer of 0.3 mil Kapton (
# WD=6 means working distance of 6cm for the detector.

def Assemble_QuadVortex(eV1):
    # quad vortex detector efficiency. eV1: fluo energy
    net=1.
    BeVortex=Material('Be', 1.85, 0.00125)
    SiO2Vortex=Material('SiO2', 2.2, 0.00001)
    SiVortex=Material('Si', 2.33, 0.035)
    BeVortex.getLa(eV1, 1)  # one Be layer in Vortex
    SiO2Vortex.getLa(eV1, 1)    # oxide layer on Si detection layer
    SiVortex.getLa(eV1, 1)  # Si detection layer, what's absorbed is counted.
    net=net*BeVortex.trans*SiO2Vortex.trans*SiVortex.absrp
    if (print2screen):
        print( '%.3f eV : BeVortex.trans=%.3e , SiO2Vortex.trans=%.3e, SiVortex.absrp=%.3e, Det_efficiency=%.3e' % (eV1, BeVortex.trans,  SiO2Vortex.trans, SiVortex.absrp, net))
    return net


def Assemble_Collimator(eV1, xHe=1, xAl=0,xKapton=0, WD=6.0, xsw=0):
    # from sample surface to detector
    # xsw=0/1/-1,  6cm, 3cm, no collimator
    # He_path depends on the collliator.
    if xHe==1:
        He_path=WD
    else:
        He_path=0.
    air_path = WD-He_path
    kapton_inside=0  # kapton inside collimator.  no collimator, no kapton_inside
    if xsw==0:  # WD60mm collimator
        kapton_inside=1                 # collimator has thin Kapton on the second aperture
        if xHe==1:
            air_path = 1.51             # 1.51cm between second aperture and Be of Vortex
        else:
            air_path=WD
        He_path = WD-air_path           # He is between sample surface to the second aperture of xrm collimator
    if xsw==1:  # WD30mm colllimator
        kapton_inside=1                 # collimator has thin Kapton on the second aperture
        if xHe==1:      # modified 11/18/2010
            He_path = 1.088   # from sample surface to Kapton cover/collimator
            air_path = WD - He_path            # 1.912cm between second aperture and Be of Vortex
            xKapton = xKapton+1         # one Kapton film used to fill He from sample surface to collimator
        else:
            air_path = WD0;     He_path=0
    air=Material('N1.56O0.48C0.03Ar0.01Kr0.000001Xe0.0000009', 0.0013, air_path)
    kapton=Material('C22H10O4N2', 1.42, 0.000762)       # 0.3mil thick
    HeGas=Material('He', 0.00009, He_path)
    AlFoil=Material('Al', 2.72, 0.00381)                # 1.5mil thick
    kaptonCollimator=Material('C22H10O4N2', 1.42, 0.000762) # 0.3mil thick
    #
    air.getLa(eV1, 1)   # 1 means number of layers here.
    kapton.getLa(eV1, xKapton)  #number of Kapton layers before collimator
    HeGas.getLa(eV1, xHe) # number of He gas layers, default=0
    AlFoil.getLa(eV1, xAl)  # number of Al foil,  default=0
    kaptonCollimator.getLa(eV1, kapton_inside)   # without collimator, no addition kapton inside
    #
    net=air.trans*HeGas.trans
    net=net*kapton.trans*AlFoil.trans*kaptonCollimator.trans
    if print2screen:
            print('%.3f eV: air.trans=%.3e, HeGas.trans=%.3e, kapton.trans=%.3e, AlFoil.trans=%.3e, kaptonCollimator.trans=%.3e, net=%.3e' %  \
            (eV1, air.trans, HeGas.trans, kapton.trans, AlFoil.trans,kaptonCollimator.trans, net))
    #print('%.3f eV: air.la=%.3e, HeGas.la=%.3e, kapton.la=%.3e' % (eV1, air.la, HeGas.la, kapton.la))
    return net


def Assemble_Detector(eV1, xHe=1, xAl=0,xKapton=0, WD=6.0, xsw=0):
    det_efficiency=Assemble_QuadVortex(eV1)
    trans2det=Assemble_Collimator(eV1, xHe, xAl,xKapton, WD, xsw)
    net=det_efficiency*trans2det
    return net



'''
----------------------------------------------------------------------------------------------------------
Combine detector, transmission, sample self-absorption, and fluorescing element distribution/concentration.
----------------------------------------------------------------------------------------------------------
'''
# this function is for XSW/XRM.
def cal_NetYield2(eV0, Atoms, xHe=0, xAl=0, xKapton=0, WD=6.0, xsw=0, WriteFile='Y' , xsect_resonant=0.0, sample=''):
    #   incident energy, list of elements, experimental conditions
    #   this one tries Ka, Kb, Lg, Lb, La, Lb
    angle0=45.; textOut=''
    if xsw!=0 and WriteFile=='Y':  print( 'XSW measurements')
    if sample=='':
        Include_SelfAbsorption='No'
    else:
        Include_SelfAbsorption='Yes'
        angle0=sample.angle/math.pi*180.
        textOut=sample.txt          # substrate concentration
    NetYield=[];    NetTrans=[]; Net=[];    out2='';	net=0.0
    text0=''
    edges      =['K',  'K',  'L1', 'L1', 'L2', 'L2', 'L2', 'L3', 'L3', 'L3']
    Fluo_lines =['Ka', 'Kb', 'Lb', 'Lg', 'Ln', 'Lb', 'Lg', 'Ll', 'La', 'Lb']
    outputfile='Elemental_Sensitivity.txt'
    if WriteFile=='Y':
        fo=open(outputfile, 'w')
        desc=' '
        if xHe==0:   desc=' not '
        if xsw==-1:
            out1='# 13IDC XRM/XSW using QuadVortex, incident x-ray energy at '+str(eV0)+' eV at '+str(angle0)+' degrees \n'
            out1+='# Helium path'+desc+'used, '+str(xAl)+' Al attenuators, '+str(xKapton+1)+' Kapton attenuators, '\
              +str(WD)+' cm working distance. \n'
        else:
            out1='# 13IDC XRM/XSW using QuadVortex + collimator,  incident x-ray energy at '+str(eV0)+' eV at '+str(angle0)+' degrees \n'
            out1+='# Helium path'+desc+'used, '+str(xAl)+' Al attenuators, '+str(xKapton)+' Kapton attenuators, '\
              +str(WD)+' cm working distance. \n'
        print( out1)
        fo.write(out1)
        if sample!='':
            for stuff in Atoms:
                text1='%6.3e %s/cm^3' % (stuff.Conc*sample.Nat1, stuff.AtSym)
                text0=text0+' '+text1
        textOut=textOut+' '+text0       # substrate concentration + other concentrations
        out1='# '+text0+'\n'
        if print2screen:
            print( out1)
        fo.write(out1)
        out1='%s\t%s\t%s   \t%s   \t%s   \t%s   \t%s\n' % ('atom', 'emit', 'emit_energy', 'yield', 'transmission', 'net_sensitivity', 'sensitivity*concentration')
        if print2screen:
            print( out1)
        fo.write(out1)
    for (ii, atom) in enumerate(Atoms):
        # MN replace:
        atnum=f1f2.AtSym2AtNum(atom.AtSym)
        # MN replace:
        temp=f1f2.AtNum2f1f2Xsect(atnum, eV0)
        xsect=temp[2]                           # photo-electric cross-section for fluorescence yield, temp[4] is total for attenuation
        con=atom.Conc
        for (nn, edge) in enumerate(edges):
            emit=Fluo_lines[nn]
            fy, emit_eV, emit_prob = fluo_yield(atom.AtSym, edge, emit, eV0)
            print(emit)
            if fy==0.0 or emit_prob==0:
                continue                        # try next item if FY=0
            else:
                if xsect_resonant!=0.0:         # use input value near edge
                    xsect=xsect_resonant	# for cross-section near absoprtion edge
                #print(xsect,fy,emit_prob)
                net_yield=xsect*fy*emit_prob    # net_yield --> cross-section, yield, emission_probability
                # net transmission --> transmission from surface through detector
## ------------------   [self-absorption]     ------------------------------
                trans_SelfAbsorp = 1.0           # for self-absorption.
                if Include_SelfAbsorption=='Yes':
                    sample.getPenetration(eV0)   # incident x-ray attenuation
                    if fy*emit_eV*emit_prob==0:  # any one of three is zero
                        break                    # skip
                    sample.getLa(emit_eV)
                    # account for incident x-ray attenuation, emitted x-ray attenuation
                    trans_SelfAbsorp = sample.scale         # for elements that are not part of substrate
                    if (atom in sample.ElemListFY):
                        if atom.tag=='substrate1':
                            trans_SelfAbsorp = sample.scale1  # for elements that are part of top substrate
                        if atom.tag=='substrate2':
                            trans_SelfAbsorp = sample.scale2  # for elements that are part of bottom substrate
## ----------------------------------------------------------------------------
            net_trans = Assemble_Detector(emit_eV, xHe, xAl,xKapton, WD, xsw)
            net = net_yield*net_trans*trans_SelfAbsorp  # elemental sensitivity
            inten = net*con  # sensitivity * concentration
            if WriteFile=='Y':
                out1='%s\t%s\t%6.1f   \t%.3e   \t%.3e   \t%.3e   \t%.3e\n' % (atom.AtSym+'_'+edge, emit, emit_eV, net_yield, net_trans, net, inten)
                fo.write(out1)
            if print2screen:
                print('%s %s %6.1f net_yield=%.3e net_trans=%.3e net=%.3e\t' % (atom.AtSym+'_'+edge, emit, emit_eV, net_yield, net_trans, net))
            #print(out1+'  %s, depth-dependent factor= %6.4f' % (atom.tag, trans_SelfAbsorp))
            if emit=='Kb' and fy!=0:        # if above K edge, don't bother trying L edges
                break
    return textOut


#def sim_spectra(eV0, Atoms, Conc, xHe=0, xAl=0, xKapton=0, WD=6.0, xsw=0, sample=''):
def sim_spectra(eV0, Atoms, xHe=0, xAl=0, xKapton=0, WD=6.0, xsw=0, sample=''):
    # sample=sample matrix with object attribues to add self-absorption effect
    # Atoms is a list with elements that have attributes AtSym, Conc, tag
    if xsw==-1:     xKapton=xKapton-1   # no collimator
    Include_SelfAbsorption='Yes'
    Print2Screen='No'
    if sample=='': Include_SelfAbsorption='No'
    if xsw!=0:  print('XSW measurements')
    xx=[]; yy=[]; tag=[];   intensity_max=-10.0; LoLimit=1e-10
    angle0=''; text1=''
    if sample!='':      # sample matrix option is used
        angle0=str(sample.angle*180./math.pi)
        angle0=angle0[:5]
        for (ii, item) in enumerate(sample.ElemListFY):  # add matrix elements to the lists
            Atoms.append(item)
    outputfile='simSpectrum_table.txt'
    fo=open(outputfile, 'w')
    out1='#incident x-ray at '+str(eV0)+' eV and '+angle0+' Deg.\n'
    if Print2Screen=='Yes': print(out1)
    fo.write(out1)
    out1='#Emission\tenergy(eV)\tintensity \n'
    if Print2Screen=='Yes': print(out1)
    fo.write(out1)
    out2='#'
    for (ix,atom) in enumerate(Atoms):
        # MN replace:
        atnum=f1f2.AtSym2AtNum(atom.AtSym)
        # con=Conc[ix]
        con=atom.Conc
        out2=out2+atom.AtSym+'['+str(con)+']   '
        for edge in ['K', 'L1', 'L2', 'L3']:
            # MN replace:
            temp=f1f2.AtNum2f1f2Xsect(atnum, eV0)
            xsect=temp[2]           # photoelectric crosssection for each element at incident x-ray, fluorescence yield
            # MN replace:
            temp=elam.use_ElamFY(atom.AtSym, edge)
            edge_eV=float(temp[2])   # absorption edge
            if eV0>edge_eV:
                fy=float(temp[3])
                for EmitLine in temp[4]:
                    EmitName=EmitLine[1]
                    emit_eV=float(EmitLine[2])
                    emit_prob=float(EmitLine[3])
                    if emit_eV<fluo_emit_min:  continue        # ignore fluorescence below fluo_emit_min (global variable)
                    name = atom.AtSym + '_'+ EmitName
                    # net transmission --> transmission from surface through detector
## ------------------   [self-absorption]     ------------------------------
                    trans_SelfAbsorp = 1.0              # for self-absorption.
                    if Include_SelfAbsorption=='Yes':
                        sample.getPenetration(eV0)      # incident x-ray attenuation
                        eV0str=str(eV0)
                        text1=' absorption_length1(%seV)= %2.2e%s \
                                absorption_length2(%seV)= %2.2e%s' % (eV0str, sample.la1*1.e4, 'microns', eV0str, sample.la2*1.e4, 'microns')
                        # text1 is added to sample.txt later
                        sample.getLa(emit_eV)           # emitted fluorescence attenuation
                        trans_SelfAbsorp = sample.scale # for elements that are not part of substrate
                        if (atom in sample.ElemListFY):
                            if atom.tag=='substrate1':
                                trans_SelfAbsorp = sample.scale1  # for elements that are part of top substrate
                            if atom.tag=='substrate2':
                                trans_SelfAbsorp = sample.scale2  # for elements that are part of bottom substrate
## ----------------------------------------------------------------------------
                    trans = Assemble_Detector(emit_eV, xHe, xAl,xKapton, WD, xsw)
                    intensity = con * fy * emit_prob * xsect * trans * trans_SelfAbsorp
                    if intensity<LoLimit: continue            # skip weak emission, arbtraray limit =1e-10.
                    if intensity>intensity_max: intensity_max=intensity
                    xx.append(emit_eV);    yy.append(intensity);   tag.append(name)
    for ix in range(len(yy)):
        yy[ix]=yy[ix]/intensity_max*100.00          # makes the strongest line to 100.0
        out1='%s\t%f\t%f \n' % (tag[ix], xx[ix], yy[ix])
        if Print2Screen=='Yes': print(out1)
        fo.write(out1)
    if Print2Screen=='Yes': print(out2)
    fo.write(out2)
    fo.close()
    out1=sim_GaussPeaks(xx, yy, det_res, eV0)        # det_res: detector resoultion for Gaussian width (global variable)
    if Include_SelfAbsorption=='Yes':
        sample.txt=sample.txt+text1                 # sample.txt is combined to output of cal_NetYield
    text=cal_NetYield2(eV0, Atoms, xHe, xAl, xKapton, WD, xsw, sample=sample)  # calculate net yield with weight-averaged emission
    print(out2)
    return text  # cal_NetYield2 output is str with number densities of elements


def sim_GaussPeaks(xx, yy, width, eV0):  #xx, yy: lists, width: a peak width, eV0: incident energy
    xline=[]; yline=[]; dX=10.0;    minX=fluo_emit_min #10eV steps, lowest-->fluo_emit_min
    amp=100                             # arbitrary multiplier to shift up spectrum
    NumOfSteps = int((eV0-minX)/dX)
    NumOfPeaks = len(xx)
    for ix in range(NumOfSteps+1):
        xline.append(0);  yline.append(0)
    for (iy,peak) in enumerate(xx):
        X0=float(peak)
        Y0=float(yy[iy])
        for ix in range(NumOfSteps+1):
            energy = minX + ix*dX
            inten = Y0*math.exp(-((energy-X0)/width)**2)*amp
            xline[ix]=energy
            yline[ix]+=inten
    outputfile='simSpectrum_plot.txt'
    fo=open(outputfile, 'w')
#    DetectorLimit=1e5                   # upperlimit for total counts to 1e5 (1e5 CPS)
    LoLimit=0.001                         # low limit for each channel
#    total=0.0
    factor=1.0
    for ix in range(NumOfSteps+1):
        if yline[ix]<LoLimit: yline[ix]=LoLimit
#        total+=yline[ix]  # add counts
#    factor=total/DetectorLimit
    for ix in range(NumOfSteps+1):
        yline[ix]=yline[ix]/factor
        out1=str(xline[ix])+'\t'+str(yline[ix])+'\n'
        fo.write(out1)
    fo.close()
    #return xline, yline


class input_param:
    def __init__(self, eV0=14000,
                 Atoms=[],
                 xHe=0.0,
                 xAl=0,
                 xKap=0,
                 WD=6.0,
                 xsw=0):
        if Atoms==[]:
            atom1=ElemFY()
            Atoms.append(atom1)
        self.eV0=eV0
        #incident x-ray energy in eV
        list1=[]; list2=[]
        for item in Atoms:
            list1.append(item.AtSym)
            list2.append(item.Conc)
        self.Atoms=list1
        #elements list
        self.Conc=list2
        #relative concentrations list
        self.xHe=xHe
        #He gas path before collimator?
        self.xAl=xAl
        #number of Al foils as attenuator
        self.xKap=xKap
        #number of Kapton foils as attenuator
        self.WD=WD
        #working distance in cm
        self.xsw=xsw
        #x-ray standing wave setup?  WD=3.0 for xsw=1



# ----------------------------------------------------------------



if __name__=='__main__':
    testing = 0
    if testing:
        #atom0='Fe'
        #emission0='Ka'
        #eV0=10000.
        #edge0='K'
        #eV1=6400.
        #mat0=Material('SiO2', 2.2, 0.001)
        #mat0.getLa(8000)
        #print(mat0.la, mat0.trans)
        #print(Assemble_QuadVortex(eV1))
        #print(Assemble_Collimator(eV1, xHe=1, xAl=0,xKapton=0, WD=6.0, xsw=0))
        # MN replace:
        matrix=SampleMatrix2('CaCO3', f1f2.nominal_density('CaCO3'), 0.001,'Fe2O3', f1f2.nominal_density('Fe2O3'), 0.001, 45.,  'all')
        eV0=7500.
        Atoms=[]
        atom=ElemFY('La', 10e-6);     Atoms.append(atom)
        atom=ElemFY('Ce', 10e-6);     Atoms.append(atom)
        atom=ElemFY('Nd', 10e-6);     Atoms.append(atom)
        for (ii, item) in enumerate(matrix.ElemListFY):
            pass
        sim_spectra(eV0, Atoms, sample=matrix)  # xKapton=-1 to remove WD60, -2 for WD30
    else:
        # March 2012 for Nov2011 beamtime
        eV0 = 10000.
        print2screen=0
        Atoms = []
        atom=ElemFY('Al', 10e-6);     Atoms.append(atom)
        atom=ElemFY('Si', 10e-6);     Atoms.append(atom)
        atom=ElemFY('Ca', 10e-6);     Atoms.append(atom)
        atom=ElemFY('Cr', 10e-6);     Atoms.append(atom)
        atom=ElemFY('Mn', 10e-6);     Atoms.append(atom)
        atom=ElemFY('Fe', 10e-6);     Atoms.append(atom)
        atom=ElemFY('Ni', 10e-6);     Atoms.append(atom)
        #sim_spectra(eV0, Atoms, xHe=1, xAl=0, xKapton=0, WD=3.0, xsw=1, sample='')
        #Assemble_QuadVortex(1486.56)
        #Assemble_Collimator(1486.56, 1, 0, 0, 3., 1)
        #
        air_path = 1.088
        He_path = 1.912
        air=Material('N1.56O0.48C0.03Ar0.01Kr0.000001Xe0.0000009', 0.0013, air_path)
        kapton=Material('C22H10O4N2', 1.42, 0.000762)       # 0.3mil thick
        HeGas=Material('He', 0.00009, He_path)
        AlFoil=Material('Al', 2.72, 0.00381)                # 1.5mil thick
        #
        from math import *
        emitE = [
            1486.4, 1557.0, 1739.6, 1837.0, 3691.1, 4013.1, 5411.6,
            5947.0, 5896.5, 6492.0, 6400.8, 7059.6, 7474.4, 8266.6
            ]
        for eV1 in emitE:
            air.getLa(eV1, 1)   # 1 means number of layers here.
            print( eV1, air.la, exp(-1./air.la))


'''
def AtSym2FY(AtSym, Shell):  #returns fluorescence yield using atomic symbol and edge(K,L1,L2,L3,M)
    out=elam.use_ElamFY(AtSym, Shell)
    #AtSym, edge, edge-energy, FY, [ [transition, emission, energy, probability],[]...]
    FY=out[3]
    return FY

class Element:
    def __init__(self, AtNum, AtWt, f1, f2, Xsection):
        self.AtNum=AtNum
        self.AtWt=AtWt              #atomic weight g/mol
        self.f1=f1                  #real part of scattering factor
        self.f2=f2                  #imaginary part of scattering factor
        self.Xsection=Xsection      #atomic crossection Barns/atom
        # f1, f2, Xsection depends on incident x-ray energy


def setElement(AtSym, shell, eV0):
    AtNum=AtSym2AtNum(AtSym)
    AtWt=AtSym2AtWt(AtSym)
    f1f2=AtNum2f1f2Xsect(AtNum, eV0)
    AtSym=Element(AtNum, AtWt, f1f2[0], f1f2[1], f1f2[2])
    return AtSym


class input_param:
    def __init__(self, eV0=14000,
                 Atoms=[],
                 xHe=0.0,
                 xAl=0,
                 xKap=0,
                 WD=6.0,
                 xsw=0):
        if Atoms==[]:
            atom1=ElemFY()
            Atoms.append(atom1)
        self.eV0=eV0
        #incident x-ray energy in eV
        list1=[]; list2=[]
        for item in Atoms:
            list1.append(item.AtSym)
            list2.append(item.Conc)
        self.Atoms=list1
        #elements list
        self.Conc=list2
        #relative concentrations list
        self.xHe=xHe
        #He gas path before collimator?
        self.xAl=xAl
        #number of Al foils as attenuator
        self.xKap=xKap
        #number of Kapton foils as attenuator
        self.WD=WD
        #working distance in cm
        self.xsw=xsw
        #x-ray standing wave setup?  WD=3.0 for xsw=1
'''



#!/usr/bin/env python
"""
Linear Combination panel
"""
import os
import time
import wx
import wx.grid as wxgrid
import numpy as np
import pickle
import base64

from functools import partial
from collections import OrderedDict

from lmfit.printfuncs import gformat
from larch import Group
from larch.math import index_of
from larch.wxlib import (BitmapButton, TextCtrl, FloatCtrl, get_icon,
                         SimpleText, pack, Button, HLine, Choice, Check,
                         NumericCombo, CEN, LEFT, Font)
from larch.io import read_csv
from larch.utils.strutils import fix_varname

from .taskpanel import TaskPanel, DataTableGrid

# plot options:
norm   = 'Normalized \u03bC(E)'
dmude  = 'd\u03bC(E)/dE'
chik   = '\u03c7(k)'
noplot = '<no plot>'
noname = '<none>'

CSV_WILDCARDS = "CSV Files(*.csv,*.dat)|*.csv*;*.dat|All files (*.*)|*.*"
MODEL_WILDCARDS = "Regression Model Files(*.regmod,*.dat)|*.regmod*;*.dat|All files (*.*)|*.*"

FitSpace_Choices = [norm, dmude, chik]
Plot_Choices = ['Mean Spectrum + Active Energies',
                'Spectra Stack',
                'Predicted External Varliable']

Regress_Choices = ['Partial Least Squares', 'LassoLars']

defaults = dict(fitspace=norm, varname='valence', xmin=-5.e5, xmax=5.e5,
                scale=True, cv_folds=None, cv_repeats=3, fit_intercept=True,
                use_lars=True, alpha=0.01)

MAX_ROWS = 1000

def make_steps(max=1, decades=8):
    steps = [1.0]
    for i in range(6):
        steps.extend([(j*10**(-(1+i))) for j in (5, 2, 1)])
    return steps

class RegressionPanel(TaskPanel):
    """Regression Panel"""
    def __init__(self, parent, controller, **kws):
        TaskPanel.__init__(self, parent, controller,
                           configname='regression_config', config=defaults,
                           title='Regression and Feature Selection', **kws)
        self.result = None
        self.save_csvfile   = 'RegressionData.csv'
        self.save_modelfile = 'Model.regmod'

    def process(self, dgroup, **kws):
        """ handle processing"""
        if self.skip_process:
            return
        self.skip_process = True
        form = self.read_form()

    def build_display(self):
        panel = self.panel
        wids = self.wids
        self.skip_process = True

        wids['fitspace'] = Choice(panel, choices=FitSpace_Choices, size=(250, -1))
        wids['fitspace'].SetStringSelection(norm)
        # wids['plotchoice'] = Choice(panel, choices=Plot_Choices,
        #                           size=(250, -1), action=self.onPlot)

        wids['method'] = Choice(panel, choices=Regress_Choices, size=(250, -1),
                                action=self.onRegressMethod)

        add_text = self.add_text

        opts = dict(digits=2, increment=1.0)

        w_xmin = self.add_floatspin('xmin', value=defaults['xmin'], **opts)
        w_xmax = self.add_floatspin('xmax', value=defaults['xmax'], **opts)
        wids['alpha'] =  NumericCombo(panel, make_steps(), fmt='%.6g',
                                      default_val=0.01, width=100)

        wids['auto_scale_pls'] = Check(panel, default=True, label='auto scale?')
        wids['auto_alpha'] = Check(panel, default=False, label='auto alpha?')

        wids['fit_intercept'] = Check(panel, default=True, label='fit intercept?')

        wids['save_csv'] = Button(panel, 'Save CSV File', size=(150, -1),
                                    action=self.onSaveCSV)
        wids['load_csv'] = Button(panel, 'Load CSV File', size=(150, -1),
                                    action=self.onLoadCSV)

        wids['save_model'] = Button(panel, 'Save Model', size=(150, -1),
                                    action=self.onSaveModel)
        wids['save_model'].Disable()

        wids['load_model'] = Button(panel, 'Load Model', size=(150, -1),
                                    action=self.onLoadModel)


        wids['train_model'] = Button(panel, 'Train Model From These Data',
                                     size=(275, -1),  action=self.onTrainModel)

        wids['fit_group'] = Button(panel, 'Predict Variable for Selected Groups',
                                   size=(275, -1), action=self.onPredictGroups)
        wids['fit_group'].Disable()


        w_cvfolds = self.add_floatspin('cv_folds', digits=0, with_pin=False,
                                       value=0, increment=1, min_val=-1)

        w_cvreps  = self.add_floatspin('cv_repeats', digits=0, with_pin=False,
                                       value=0, increment=1, min_val=-1)

        w_ncomps  = self.add_floatspin('ncomps', digits=0, with_pin=False,
                                       value=3, increment=1, min_val=1)

        wids['varname'] = wx.TextCtrl(panel, -1, 'valence', size=(150, -1))
        wids['stat1'] =  SimpleText(panel, ' - - - ')
        wids['stat2'] =  SimpleText(panel, ' - - - ')


        collabels = [' File /Group Name ', 'External Value',
                     'Predicted Value']
        colsizes = [300, 120, 120]
        coltypes = ['str', 'float:12,4', 'float:12,4']
        coldefs  = ['', 0.0, 0.0]

        wids['table'] = DataTableGrid(panel, nrows=MAX_ROWS,
                                      collabels=collabels,
                                      datatypes=coltypes,
                                      defaults=coldefs,
                                      colsizes=colsizes)

        wids['table'].SetMinSize((625, 175))

        wids['use_selected'] = Button(panel, 'Use Selected Groups',
                                      size=(150, -1),  action=self.onFillTable)

        panel.Add(SimpleText(panel, 'Feature Regression, Model Selection',
                             **self.titleopts), dcol=4)
        add_text('Array to Use: ', newrow=True)
        panel.Add(wids['fitspace'], dcol=4)

        # add_text('Plot : ', newrow=True)
        # panel.Add(wids['plotchoice'], dcol=3)
        add_text('Fit Energy Range: ')
        panel.Add(w_xmin)
        add_text(' : ', newrow=False)
        panel.Add(w_xmax, dcol=3)
        add_text('Regression Method:')
        panel.Add(wids['method'], dcol=4)
        add_text('PLS # components: ')
        panel.Add(w_ncomps)
        panel.Add(wids['auto_scale_pls'], dcol=2)
        add_text('Lasso Alpha: ')
        panel.Add(wids['alpha'])
        panel.Add(wids['auto_alpha'], dcol=2)
        panel.Add(wids['fit_intercept'])
        wids['alpha'].Disable()
        wids['auto_alpha'].Disable()
        wids['fit_intercept'].Disable()

        add_text('Cross Validation: ')
        add_text(' # folds, # repeats: ', newrow=False)
        panel.Add(w_cvfolds, dcol=2)
        panel.Add(w_cvreps)

        panel.Add(HLine(panel, size=(600, 2)), dcol=6, newrow=True)

        add_text('Build Model: ', newrow=True)
        panel.Add(wids['use_selected'],   dcol=2)
        add_text('Attribute Name: ', newrow=False)
        panel.Add(wids['varname'], dcol=4)

        add_text('Read/Save Data: ', newrow=True)
        panel.Add(wids['load_csv'], dcol=3)
        panel.Add(wids['save_csv'], dcol=2)

        panel.Add(wids['table'], newrow=True, dcol=5) # , drow=3)

        panel.Add(HLine(panel, size=(550, 2)), dcol=5, newrow=True)
        panel.Add((5, 5), newrow=True)
        add_text('Train Model : ')
        panel.Add(wids['train_model'], dcol=3)
        panel.Add(wids['load_model'])

        add_text('Use This Model : ')
        panel.Add(wids['fit_group'], dcol=3)
        panel.Add(wids['save_model'])
        add_text('Fit Statistics : ')
        panel.Add(wids['stat1'], dcol=4)
        panel.Add((5, 5), newrow=True)
        panel.Add(wids['stat2'], dcol=4)
        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((10, 10), 0, LEFT, 3)
        sizer.Add(panel, 1, LEFT, 3)
        pack(self, sizer)
        self.skip_process = False

    def onRegressMethod(self, evt=None):
        meth = self.wids['method'].GetStringSelection()
        use_lasso = meth.lower().startswith('lasso')
        self.wids['alpha'].Enable(use_lasso)
        self.wids['auto_alpha'].Enable(use_lasso)
        self.wids['fit_intercept'].Enable(use_lasso)
        self.wids['fit_intercept'].Enable(use_lasso)
        self.wids['auto_scale_pls'].Enable(not use_lasso)
        self.wids['ncomps'].Enable(not use_lasso)


    def fill_form(self, dgroup):
        opts = self.get_config(dgroup)
        self.dgroup = dgroup
        if isinstance(dgroup, Group):
            d_emin = min(dgroup.energy)
            d_emax = max(dgroup.energy)
            if opts['xmin'] < d_emin:
                opts['xmin'] = -40 + int(dgroup.e0/10.0)*10
            if opts['xmax'] > d_emax:
                opts['xmax'] =  110 + int(dgroup.e0/10.0)*10

        self.skip_process = True
        wids = self.wids
        for attr in ('xmin', 'xmax', 'alpha'):
            val = opts.get(attr, None)
            if val is not None:
                if attr == 'alpha':
                    val = "%.6g" % val
                wids[attr].SetValue(val)

        for attr in ('fitspace',):
            if attr in opts:
                wids[attr].SetStringSelection(opts[attr])

        self.skip_process = False

    def onFillTable(self, event=None):
        selected_groups = self.controller.filelist.GetCheckedStrings()
        varname = fix_varname(self.wids['varname'].GetValue())
        predname = varname + '_predicted'
        grid_data = []
        for fname in self.controller.filelist.GetCheckedStrings():
            gname = self.controller.file_groups[fname]
            grp = self.controller.get_group(gname)
            grid_data.append([fname, getattr(grp, varname, 0.0),
                              getattr(grp, predname, 0.0)])

        self.wids['table'].table.data = grid_data
        self.wids['table'].table.View.Refresh()

    def onTrainModel(self, event=None):
        opts = self.read_form()
        varname = opts['varname']
        predname = varname + '_predicted'

        grid_data = self.wids['table'].table.data
        groups = []
        for fname, yval, pval in grid_data:
            gname = self.controller.file_groups[fname]
            grp = self.controller.get_group(gname)
            setattr(grp, varname, yval)
            setattr(grp, predname, pval)
            groups.append(gname)

        cmds = ['# train linear regression model',
               'training_groups = [%s]' % ', '.join(groups)]

        copts = ["varname='%s'" % varname,
                 "xmin=%.4f" % opts['xmin'],
                 "xmax=%.4f" % opts['xmax']]

        arrname = 'norm'
        if opts['fitspace'] == dmude:
            arrname = 'dmude'
        elif opts['fitspace'] == chik:
            arrname = 'chi'
        copts.append("arrayname='%s'" % arrname)

        self.method = 'pls'
        if opts['method'].lower().startswith('lasso'):
            self.method = 'lasso'
            if opts['auto_alpha']:
                copts.append('alpha=None')
            else:
                copts.append('alpha=%s' % opts['alpha'])
            copts.append('use_lars=%s' % repr('lars' in opts['method'].lower()))
            copts.append('fit_intercept=%s' % repr(opts['fit_intercept']))
        else:
            copts.append('ncomps=%d' % opts['ncomps'])
            copts.append('scale=%s' % repr(opts['auto_scale_pls']))

        copts = ', '.join(copts)
        cmds.append("reg_model = %s_train(training_groups, %s)" %
                    (self.method, copts))
        self.larch_eval('\n'.join(cmds))
        reg_model = self.larch_get('reg_model')
        if reg_model is not None:
            self.write_message('Regression Model trained: %s' % opts['method'])
            rmse_cv = reg_model.rmse_cv
            if rmse_cv is not None:
                rmse_cv = "%.4f" % rmse_cv
            stat = "RMSE_CV = %s, RMSE = %.4f" % (rmse_cv, reg_model.rmse)
            self.wids['stat1'].SetLabel(stat)
            if self.method == 'lasso':
                stat = "Alpha = %.4f, %d active components"
                self.wids['stat2'].SetLabel(stat % (reg_model.alpha,
                                                    len(reg_model.active)))

                if opts['auto_alpha']:
                    self.wids['alpha'].add_choice(reg_model.alpha)

            else:
                self.wids['stat2'].SetLabel('- - - ')

            for i, row in enumerate(grid_data):
                grid_data[i] = [row[0], row[1], reg_model.ypred[i]]
            self.wids['table'].table.data = grid_data
            self.wids['table'].table.View.Refresh()

            if reg_model.cv_folds not in (0, None):
                self.wids['cv_folds'].SetValue(reg_model.cv_folds)
            if reg_model.cv_repeats not in (0, None):
                self.wids['cv_repeats'].SetValue(reg_model.cv_repeats)

            self.wids['save_model'].Enable()
            self.wids['fit_group'].Enable()

            wx.CallAfter(self.onPlotModel, model=reg_model)

    def onPredictGroups(self, event=None):
        opts = self.read_form()
        varname = opts['varname'] + '_predicted'

        grid_data = self.wids['table'].table.data
        gent = {}
        if len(grid_data[0][0].strip()) == 0:
            grid_data = []
        else:
            for i, row in enumerate(grid_data):
                gent[row[0]] = i

        for fname in self.controller.filelist.GetCheckedStrings():
            gname = self.controller.file_groups[fname]
            grp   = self.controller.get_group(gname)
            extval = getattr(grp, opts['varname'], 0)
            cmd = "%s.%s = %s_predict(%s, reg_model)" % (gname, varname,
                                                         self.method, gname)
            self.larch_eval(cmd)
            val = self.larch_get('%s.%s' % (gname, varname))
            if fname in gent:
                grid_data[gent[fname]][2] = val
            else:
                grid_data.append([fname, extval, val])
            self.wids['table'].table.data = grid_data
            self.wids['table'].table.View.Refresh()

    def onSaveModel(self, event=None):
        try:
            reg_model = self.larch_get('reg_model')
        except:
            reg_model = None
        if reg_model is None:
            self.write_message('Cannot Save Regression Model')
            return

        dlg = wx.FileDialog(self, message="Save Regression Model",
                            defaultDir=os.getcwd(),
                            defaultFile=self.save_modelfile,
                            wildcard=MODEL_WILDCARDS,
                            style=wx.FD_SAVE)
        fname = None
        if dlg.ShowModal() == wx.ID_OK:
            fname = dlg.GetPath()
        dlg.Destroy()
        if fname is None:
            return
        self.save_modelfile = os.path.split(fname)[1]
        text = str(base64.b64encode(pickle.dumps(reg_model)), 'utf-8')
        with open(fname, 'w') as fh:
            fh.write("%s\n" % text)
        fh.flush()
        fh.close()
        self.write_message('Wrote Regression Model to %s ' % fname)

    def onLoadModel(self, event=None):
        dlg = wx.FileDialog(self, message="Load Regression Model",
                            defaultDir=os.getcwd(),
                            wildcard=MODEL_WILDCARDS, style=wx.FD_OPEN)

        fname = None
        if dlg.ShowModal() == wx.ID_OK:
            fname = dlg.GetPath()
        dlg.Destroy()
        if fname is None:
            return
        self.save_modelfile = os.path.split(fname)[1]
        with open(fname, 'r') as fh:
            text = fh.read()

        reg_model = pickle.loads(base64.b64decode(bytes(text, 'utf-8')))
        self.controller.symtable.reg_model = reg_model
        self.write_message('Read Regression Model from %s ' % fname)
        self.wids['fit_group'].Enable()


    def onLoadCSV(self, event=None):
        dlg = wx.FileDialog(self, message="Load CSV Data File",
                            defaultDir=os.getcwd(),
                            wildcard=CSV_WILDCARDS, style=wx.FD_OPEN)

        fname = None
        if dlg.ShowModal() == wx.ID_OK:
            fname = dlg.GetPath()
        dlg.Destroy()
        if fname is None:
            return

        self.save_csvfile = os.path.split(fname)[1]
        varname = fix_varname(self.wids['varname'].GetValue())
        csvgroup = read_csv(fname)
        script = []
        grid_data = []
        for sname, yval in zip(csvgroup.col_01, csvgroup.col_02):
            if sname.startswith('#'):
                continue
            if sname in self.controller.file_groups:
                gname = self.controller.file_groups[sname]
                script.append('%s.%s = %f' % (gname, varname, yval))
                grid_data.append([sname, yval, 0])

        self.larch_eval('\n'.join(script))
        self.wids['table'].table.data = grid_data
        self.wids['table'].table.View.Refresh()
        self.write_message('Read CSV File %s ' % fname)

    def onSaveCSV(self, event=None):
        dlg = wx.FileDialog(self, message="Save CSV Data File",
                            defaultDir=os.getcwd(),
                            defaultFile=self.save_csvfile,
                            wildcard=FILE_WILDCARDS,
                            style=wx.FD_SAVE)
        fname = None
        if dlg.ShowModal() == wx.ID_OK:
            fname = dlg.GetPath()
        dlg.Destroy()
        if fname is None:
            return
        self.save_csvfile = os.path.split(fname)[1]

        buff = []
        for  row in self.wids['table'].table.data:
            buff.append("%s, %s, %s" % (row[0], gformat(row[1]), gformat(row[2])))
        buff.append('')
        with open(fname, 'w') as fh:
            fh.write('\n'.join(buff))
        self.write_message('Wrote CSV File %s ' % fname)

    def onPlotModel(self, event=None, model=None):
        opts = self.read_form()
        if model is None:
            return

        ppanel = self.controller.get_display(win=1).panel
        viewlims = ppanel.get_viewlimits()
        plotcmd = ppanel.plot

        d_ave = model.spectra.mean(axis=0)
        d_std = model.spectra.std(axis=0)
        ymin, ymax = (d_ave-d_std).min(), (d_ave+d_std).max()
        if self.method == 'lasso':
            active_coefs = (model.coefs[model.active])
            active_coefs = active_coefs/max(abs(active_coefs))
            ymin = min(active_coefs.min(), ymin)
            ymax = max(active_coefs.max(), ymax)

        else:
            ymin = min(model.coefs.min(), ymin)
            ymax = max(model.coefs.max(), ymax)

        ymin = ymin - 0.02*(ymax-ymin)
        ymax = ymax + 0.02*(ymax-ymin)


        title = '%s Regression results' % (self.method.upper())

        ppanel.plot(model.x, d_ave, win=1, title=title,
                    label='mean spectra', xlabel='Energy (eV)',
                    ylabel=opts['fitspace'], show_legend=True,
                    ymin=ymin, ymax=ymax)
        ppanel.axes.fill_between(model.x, d_ave-d_std, d_ave+d_std,
                                 color='#1f77b433')
        if self.method == 'lasso':
            ppanel.axes.bar(model.x[model.active], active_coefs,
                            1.0, color='#9f9f9f88',
                            label='coefficients')
        else:
            _, ncomps = model.coefs.shape
            for i in range(ncomps):
                ppanel.oplot(model.x, model.coefs[:, i], label='coef %d' % (i+1))

        ppanel.canvas.draw()

        ngoups = len(model.groupnames)
        indices = np.arange(len(model.groupnames))
        diff = model.ydat - model.ypred
        sx = np.argsort(model.ydat)

        ppanel = self.controller.get_display(win=2).panel

        ppanel.plot(model.ydat[sx], indices, xlabel='valence',
                    label='experimental', linewidth=0, marker='o',
                    markersize=8, win=2, new=True, title=title)

        ppanel.oplot(model.ypred[sx], indices, label='predicted',
                    labelfontsize=7, markersize=6, marker='o',
                    linewidth=0, show_legend=True, new=False)

        ppanel.axes.barh(indices, diff[sx], 0.5, color='#9f9f9f88')
        ppanel.axes.set_yticks(indices)
        ppanel.axes.set_yticklabels([model.groupnames[o] for o in sx])
        ppanel.conf.auto_margins = False
        ppanel.conf.set_margins(left=0.35, right=0.05, bottom=0.15, top=0.1)
        ppanel.canvas.draw()


    def onCopyParam(self, name=None, evt=None):
        print("on Copy Param")
        conf = self.get_config()



import time
import os
import platform
from functools import partial
from collections import OrderedDict

import numpy as np
np.seterr(all='ignore')

import wx
import wx.grid as wxgrid

from larch import Group
from larch.wxlib import (BitmapButton, SetTip, GridPanel, FloatCtrl,
                         FloatSpin, FloatSpinWithPin, get_icon, SimpleText,
                         pack, Button, HLine, Choice, Check, MenuItem,
                         GUIColors, CEN, LEFT, FRAMESTYLE, Font, FileSave,
                         FileOpen, FONTSIZE)

from larch.wxlib.plotter import last_cursor_pos
from larch.utils import group2dict

LEFT = wx.ALIGN_LEFT
CEN |=  wx.ALL

def autoset_fs_increment(wid, value):
    """set increment for floatspin to be
    1, 2, or 5 x 10^(integer) and ~0.02 X current value
    """
    if abs(value) < 1.e-20:
        return
    ndig = int(1-round(np.log10(abs(value*0.5))))
    wid.SetDigits(ndig+1)
    c, inc = 0, 10.0**(-ndig)
    while (inc/abs(value) > 0.02):
        scale = 0.5 if (c % 2 == 0) else 0.4
        inc *= scale
        c += 1
    wid.SetIncrement(inc)

class DataTable(wxgrid.GridTableBase):
    def __init__(self, nrows=50, collabels=['a', 'b'],
                 datatypes=['str', 'float:12,4'],
                 defaults=[None, None]):

        wxgrid.GridTableBase.__init__(self)

        self.ncols = len(collabels)
        self.nrows = nrows
        self.colLabels = collabels
        self.dataTypes = []
        for i, d in enumerate(datatypes):
            if d.lower().startswith('str'):
                self.dataTypes.append(wxgrid.GRID_VALUE_STRING)
                defval = ''
            elif d.lower().startswith('float:'):
                xt, opt = d.split(':')
                self.dataTypes.append(wxgrid.GRID_VALUE_FLOAT+':%s' % opt)
                defval = 0.0
            if defaults[i] is None:
                defaults[i] = defval

        self.data = []
        for i in range(self.nrows):
            self.data.append(defaults)

    def GetNumberRows(self):
        return self.nrows

    def GetNumberCols(self):
        return self.ncols

    def GetValue(self, row, col):
        try:
            return self.data[row][col]
        except IndexError:
            return ''

    def SetValue(self, row, col, value):
        self.data[row][col] = value

    def GetColLabelValue(self, col):
        return self.colLabels[col]

    def GetRowLabelValue(self, row):
        return "%d" % (row+1)

    def GetTypeName(self, row, col):
        return self.dataTypes[col]

    def CanGetValueAs(self, row, col, typeName):
        colType = self.dataTypes[col].split(':')[0]
        if typeName == colType:
            return True
        else:
            return False

    def CanSetValueAs(self, row, col, typeName):
        return self.CanGetValueAs(row, col, typeName)

class DataTableGrid(wxgrid.Grid):
    def __init__(self, parent, nrows=50, rowlabelsize=35, collabels=['a', 'b'],
                 datatypes=['str', 'float:12,4'],
                 defaults=[None, None],
                 colsizes=[200, 100]):

        wxgrid.Grid.__init__(self, parent, -1)

        self.table = DataTable(nrows=nrows, collabels=collabels,
                                datatypes=datatypes, defaults=defaults)

        self.SetTable(self.table, True)
        self.SetRowLabelSize(rowlabelsize)
        self.SetMargins(10, 10)
        self.EnableDragRowSize()
        self.EnableDragColSize()
        self.AutoSizeColumns(False)
        for i, csize in enumerate(colsizes):
            self.SetColSize(i, csize)

        self.Bind(wxgrid.EVT_GRID_CELL_LEFT_DCLICK, self.OnLeftDClick)

    def OnLeftDClick(self, evt):
        if self.CanEnableCellControl():
            self.EnableCellEditControl()


class TaskPanel(wx.Panel):
    """generic panel for main tasks.
    meant to be subclassed
    """
    def __init__(self, parent, controller, title='Generic Panel',
                 configname='task_config', config=None, **kws):
        wx.Panel.__init__(self, parent, -1, size=(550, 625), **kws)
        self.parent = parent
        self.controller = controller
        self.larch = controller.larch
        self.title = title
        self.configname = configname
        if config is not None:
            self.set_defaultconfig(config)
        self.wids = {}
        self.subframes = {}
        self.SetFont(Font(FONTSIZE))
        self.titleopts = dict(font=Font(FONTSIZE+2), colour='#AA0000')

        self.panel = GridPanel(self, ncols=7, nrows=10, pad=2, itemstyle=LEFT)
        self.panel.sizer.SetVGap(5)
        self.panel.sizer.SetHGap(5)
        self.skip_process = True
        self.skip_plotting = False
        self.build_display()
        self.skip_process = False

    def show_subframe(self, name, frameclass, **opts):
        shown = False
        if name in self.subframes:
            try:
                self.subframes[name].Raise()
                shown = True
            except:
                del self.subframes[name]
        if not shown:
            self.subframes[name] = frameclass(self, **opts)

    def onPanelExposed(self, **kws):
        # called when notebook is selected
        fname = self.controller.filelist.GetStringSelection()
        if fname in self.controller.file_groups:
            gname = self.controller.file_groups[fname]
            dgroup = self.controller.get_group(gname)
            self.fill_form(dgroup)
            self.process(dgroup=dgroup)

    def write_message(self, msg, panel=0):
        self.controller.write_message(msg, panel=panel)

    def larch_eval(self, cmd):
        """eval"""
        self.controller.larch.eval(cmd)

    def larch_get(self, sym):
        """get value from larch symbol table"""
        return self.controller.larch.symtable.get_symbol(sym)

    def build_display(self):
        """build display"""

        self.panel.Add(SimpleText(self.panel, self.title, **titleopts),
                       dcol=7)
        self.panel.Add(SimpleText(self.panel, ' coming soon....'),
                       dcol=7, newrow=True)
        self.panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.panel, 1, wx.LEFT|wx.CENTER, 3)
        pack(self, sizer)

    def set_defaultconfig(self, config):
        """set the default configuration for this session"""
        conf = self.controller.larch.symtable._sys.xas_viewer
        setattr(conf, self.configname, {key:val for key, val in config.items()})

    def get_defaultconfig(self):
        """get the default configuration for this session"""
        conf = self.controller.larch.symtable._sys.xas_viewer
        defconf = getattr(conf, self.configname, {})
        return {key:val for key, val in defconf.items()}

    def get_config(self, dgroup=None):
        """get and set processing configuration for a group"""
        if dgroup is None:
            dgroup = self.controller.get_group()
        conf = getattr(dgroup, self.configname, self.get_defaultconfig())
        if dgroup is not None:
            setattr(dgroup, self.configname, conf)
        return conf

    def update_config(self, config, dgroup=None):
        """set/update processing configuration for a group"""
        if dgroup is None:
            dgroup = self.controller.get_group()
        conf = getattr(dgroup, self.configname, self.get_defaultconfig())
        conf.update(config)
        if dgroup is not None:
            setattr(dgroup, self.configname, conf)

    def fill_form(self, dat):
        if isinstance(dat, Group):
            dat = group2dict(dat)

        for name, wid in self.wids.items():
            if isinstance(wid, FloatCtrl) and name in dat:
                wid.SetValue(dat[name])

    def read_form(self):
        "read for, returning dict of values"
        dgroup = self.controller.get_group()
        form_opts = {'groupname': dgroup.groupname}
        for name, wid in self.wids.items():
            val = None
            for method in ('GetValue', 'GetStringSelection', 'IsChecked',
                           'GetLabel'):
                meth = getattr(wid, method, None)
                if callable(meth):
                    try:
                        val = meth()
                    except TypeError:
                        pass
                if val is not None:
                    break
            form_opts[name] = val
        return form_opts

    def process(self, dgroup=None, **kws):
        """override to handle data process step"""
        if self.skip_process:
            return
        self.skip_process = True

    def add_text(self, text, dcol=1, newrow=True):
        self.panel.Add(SimpleText(self.panel, text),
                       dcol=dcol, newrow=newrow)

    def add_floatspin(self, name, value, with_pin=True, relative_e0=False,
                      **kws):
        """create FloatSpin with Pin button for onSelPoint"""
        if with_pin:
            pin_action = partial(self.onSelPoint, opt=name,
                                 relative_e0=relative_e0)
            fspin, bb = FloatSpinWithPin(self.panel, value=value,
                                         pin_action=pin_action, **kws)
        else:
            fspin = FloatSpin(self.panel, value=value, **kws)
            bb = (1, 1)

        self.wids[name] = fspin
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(fspin)
        sizer.Add(bb)
        return sizer

    def onPlot(self, evt=None):
        pass

    def onPlotOne(self, evt=None, dgroup=None, **kws):
        pass

    def onPlotSel(self, evt=None, groups=None, **kws):
        pass

    def onSelPoint(self, evt=None, opt='__', relative_e0=False, win=None):
        """
        get last selected point from a specified plot window
        and fill in the value for the widget defined by `opt`.

        by default it finds the latest cursor position from the
        cursor history of the first 20 plot windows.
        """
        if opt not in self.wids:
            return None

        _x, _y = last_cursor_pos(win=win, _larch=self.larch)

        if _x is not None:
            if relative_e0 and 'e0' in self.wids and opt != 'e0':
                _x -= self.wids['e0'].GetValue()
            self.wids[opt].SetValue(_x)
            cb = getattr(self, 'onProcess', None)
            if callable(cb):
                cb(event=None)



#!/usr/bin/env python
"""
utilities for XRF display
"""
import copy
from functools import partial
from collections import OrderedDict
import time
import numpy as np

import wx
import wx.lib.colourselect  as csel
import wx.lib.scrolledpanel as scrolled
from wxutils import (SimpleText, FloatCtrl, Choice, Font, pack, Button,
                     Check, HyperText, HLine, GridPanel, CEN, LEFT, RIGHT)

from wxmplot.colors import hexcolor
from xraydb import xray_line
from ..xrf import (xrf_calib_fitrois, xrf_calib_init_roi,
                   xrf_calib_compute, xrf_calib_apply)

# Group used to hold MCA data
XRFGROUP   = '_xrfdata'
MAKE_XRFGROUP_CMD = "%s = group(__doc__='MCA/XRF data groups', _mca='', _mca2='')" % XRFGROUP
XRFRESULTS_GROUP   = '_xrfresults'
MAKE_XRFRESULTS_GROUP = "_xrfresults = []"

def mcaname(i):
    return "mca{:03d}".format(i)

def next_mcaname(_larch):
    xrfgroup = _larch.symtable.get_group(XRFGROUP)
    i, exists = 1, True
    name = "mca{:03d}".format(i)
    while hasattr(xrfgroup, name):
        i += 1
        name = "mca{:03d}".format(i)
    return name


class XRFCalibrationFrame(wx.Frame):
    def __init__(self, parent, mca, size=(-1, -1), callback=None):
        self.mca = mca
        self.callback = callback
        wx.Frame.__init__(self, parent, -1, 'Calibrate MCA',
                          size=size, style=wx.DEFAULT_FRAME_STYLE)

        opanel = scrolled.ScrolledPanel(self)
        osizer = wx.BoxSizer(wx.VERTICAL)
        panel = GridPanel(opanel)
        self.calib_updated = False
        panel.AddText("Calibrate MCA Energy (Energies in eV)",
                      colour='#880000', dcol=7)
        panel.AddText("ROI", newrow=True, style=CEN)
        panel.AddText("Predicted", style=CEN)
        panel.AddText("Peaks with Current Calibration", dcol=3, style=CEN)
        panel.AddText("Peaks with Refined Calibration", dcol=3, style=CEN)

        panel.AddText("Name", newrow=True, style=CEN)
        panel.AddText("Energy", style=CEN)
        panel.AddText("Center", style=CEN)
        panel.AddText("Difference", style=CEN)
        panel.AddText("FWHM", style=CEN)
        panel.AddText("Center", style=CEN)
        panel.AddText("Difference", style=CEN)
        panel.AddText("FWHM", style=CEN)
        panel.AddText("Use? ", style=CEN)

        panel.Add(HLine(panel, size=(700, 3)),  dcol=9, newrow=True)
        self.wids = []

        # find ROI peak positions
        self.init_wids = {}
        for roi in self.mca.rois:
            eknown, ecen, fwhm = 1, 1, 1

            words = roi.name.split()
            elem = words[0].title()
            family = 'ka'
            if len(words) > 1:
                family = words[1]
            try:
                eknown = xray_line(elem, family).energy/1000.0
            except (AttributeError, ValueError):
                eknown = 0.0001
            mid = (roi.right + roi.left)/2
            wid = (roi.right - roi.left)/2
            ecen = mid * mca.slope + mca.offset
            fwhm = 2.354820 * wid * mca.slope

            diff = ecen - eknown
            name = ('   ' + roi.name+' '*10)[:10]
            opts = {'style': CEN, 'size':(75, -1)}
            w_name = SimpleText(panel, name,   **opts)
            w_pred = SimpleText(panel, "% .1f" % (1000*eknown), **opts)
            w_ccen = SimpleText(panel, "% .1f" % (1000*ecen),   **opts)
            w_cdif = SimpleText(panel, "% .1f" % (1000*diff),   **opts)
            w_cwid = SimpleText(panel, "% .1f" % (1000*fwhm),   **opts)
            w_ncen = SimpleText(panel, "-----",         **opts)
            w_ndif = SimpleText(panel, "-----",         **opts)
            w_nwid = SimpleText(panel, "-----",         **opts)
            w_use  = Check(panel, label='', size=(40, -1), default=0)
            panel.Add(w_name, style=LEFT, newrow=True)
            panel.AddMany((w_pred, w_ccen, w_cdif, w_cwid,
                           w_ncen, w_ndif, w_nwid, w_use))
            self.init_wids[roi.name] = [False, w_pred, w_ccen, w_cdif, w_cwid, w_use]
            self.wids.append((roi.name, eknown, ecen, w_ncen, w_ndif, w_nwid, w_use))

        panel.Add(HLine(panel, size=(700, 3)),  dcol=9, newrow=True)
        offset = 1000.0*self.mca.offset
        slope  = 1000.0*self.mca.slope
        panel.AddText("Current Calibration:",   dcol=2, newrow=True)
        panel.AddText("offset(eV):")
        panel.AddText("%.3f" % (offset), dcol=1, style=RIGHT)
        panel.AddText("slope(eV/chan):")
        panel.AddText("%.3f" % (slope),  dcol=1, style=RIGHT)

        panel.AddText("Refined Calibration:", dcol=2, newrow=True)
        self.new_offset = FloatCtrl(panel, value=offset, precision=3,
                                   size=(80, -1))
        self.new_slope = FloatCtrl(panel, value=slope,  precision=3,
                                   size=(80, -1))
        panel.AddText("offset(eV):")
        panel.Add(self.new_offset,    dcol=1, style=RIGHT)
        panel.AddText("slope(eV/chan):")
        panel.Add(self.new_slope,     dcol=1, style=RIGHT)

        self.calib_btn = Button(panel, 'Compute Calibration',
                                size=(175, -1), action=self.onCalibrate)
        self.calib_btn.Disable()
        panel.Add(self.calib_btn, dcol=3, newrow=True)
        panel.Add(Button(panel, 'Use New Calibration',
                         size=(175, -1), action=self.onUseCalib),
                  dcol=3, style=RIGHT)
        panel.Add(Button(panel, 'Done',
                         size=(125, -1), action=self.onClose),
                  dcol=2, style=RIGHT)
        panel.pack()
        a = panel.GetBestSize()
        self.SetSize((a[0]+25, a[1]+50))
        osizer.Add(panel)
        pack(opanel, osizer)
        opanel.SetupScrolling()
        self.Show()
        self.Raise()
        self.init_proc = False
        self.init_t0 = time.time()
        self.init_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.onInitTimer, self.init_timer)
        self.init_timer.Start(2)

    def onInitTimer(self, evt=None):
        """initial calibration"""
        if self.init_proc:
            # print("skipping in init_proc...")
            return
        nextroi = None
        if time.time() - self.init_t0 > 20:
            self.init_timer.Stop()
            self.init_proc = False
            self.calib_btn.Enable()
        for roiname, wids in self.init_wids.items():
            if not wids[0]:
                nextroi = roiname
                break

        if nextroi is None:
            self.init_timer.Stop()
            self.init_proc = False
            self.calib_btn.Enable()
        else:
            self.init_proc = True
            xrf_calib_init_roi(self.mca, roiname)
            s, w_pred, w_ccen, w_cdif, w_cwid, w_use = self.init_wids[roiname]
            if roiname in self.mca.init_calib:
                eknown, ecen, fwhm, amp, fit = self.mca.init_calib[roiname]

                diff = ecen - eknown
                opts = {'style': CEN, 'size':(75, -1)}
                w_pred.SetLabel("% .1f" % (1000*eknown))
                w_ccen.SetLabel("% .1f" % (1000*ecen))
                w_cdif.SetLabel("% .1f" % (1000*diff))
                w_cwid.SetLabel("% .1f" % (1000*fwhm))
                if fwhm > 0.001 and fwhm < 2.00 and fit is not None:
                    w_use.SetValue(1)

            self.init_wids[roiname][0] = True
            self.init_proc = False


    def onCalibrate(self, event=None):
        x, y = [], []
        mca = self.mca
        # save old calib
        old_calib  =  mca.offset, mca.slope
        init_calib =  copy.deepcopy(mca.init_calib)
        for roiname, eknown, ecen, w_ncen, w_ndif, w_nwid, w_use in self.wids:
            if not w_use.IsChecked():
                mca.init_calib.pop(roiname)
            w_ncen.SetLabel("-----")
            w_ndif.SetLabel("-----")
            w_nwid.SetLabel("-----")


        xrf_calib_compute(mca, apply=True)
        offset, slope = mca.new_calib
        self.calib_updated = True
        self.new_offset.SetValue("% .3f" % (1000*offset))
        self.new_slope.SetValue("% .3f" % (1000*slope))

        # find ROI peak positions using this new calibration
        xrf_calib_fitrois(mca)
        for roi in self.mca.rois:
            try:
                eknown, ecen, fwhm, amp, fit = mca.init_calib[roi.name]
            except:
                continue
            diff  = ecen - eknown
            for roiname, eknown, ocen, w_ncen, w_ndif, w_nwid, w_use in self.wids:
                if roiname == roi.name and w_use.IsChecked():
                    w_ncen.SetLabel("%.1f" % (1000*ecen))
                    w_ndif.SetLabel("% .1f" % (1000*diff))
                    w_nwid.SetLabel("%.1f" % (1000*fwhm))
                    break


        # restore calibration to old values until new values are accepted
        xrf_calib_apply(mca, offset=old_calib[0], slope=old_calib[1])
        mca.init_calib = init_calib

        tsize = self.GetSize()
        self.SetSize((tsize[0]+1, tsize[1]))
        self.SetSize((tsize[0], tsize[1]))

    def onUseCalib(self, event=None):
        mca = self.mca
        offset = 0.001*float(self.new_offset.GetValue())
        slope  = 0.001*float(self.new_slope.GetValue())
        mca.new_calib = offset, slope
        xrf_calib_apply(mca, offset=offset, slope=slope)
        if callable(self.callback):
            self.callback(mca)
        self.Destroy()

    def onClose(self, event=None):
        self.Destroy()


class ColorsFrame(wx.Frame):
    """settings frame for XRFDisplay"""
    def __init__(self, parent, size=(400, 300), **kws):
        self.parent = parent
        conf = parent.conf
        kws['style'] = wx.DEFAULT_FRAME_STYLE
        wx.Frame.__init__(self, parent, -1, size=size,
                          title='XRF Color Settings', **kws)

        panel = GridPanel(self)
        panel.SetFont(Font(11))
        def add_color(panel, name):
            cval = hexcolor(getattr(conf, name))
            c = csel.ColourSelect(panel,  -1, "", cval, size=(35, 25))
            c.Bind(csel.EVT_COLOURSELECT, partial(self.onColor, item=name))
            return c

        def scolor(txt, attr, **kws):
            panel.AddText(txt, size=(130, -1), style=LEFT, font=Font(11), **kws)
            panel.Add(add_color(panel, attr),  style=LEFT)

        panel.AddText('    XRF Display Colors', dcol=4, colour='#880000')
        panel.Add(HLine(panel, size=(400, 3)),  dcol=4, newrow=True)
        scolor(' Main Spectra:',        'spectra_color', newrow=True)
        scolor(' Background Spectra:',      'spectra2_color')
        scolor(' ROIs:',                'roi_color',     newrow=True)
        scolor(' ROI Fill:',            'roi_fillcolor')
        scolor(' Cursor:',              'marker_color',  newrow=True)
        scolor(' XRF Background:',      'bgr_color')
        scolor(' Major X-ray Lines:',   'major_elinecolor', newrow=True)
        scolor(' Minor X-ray Lines:',   'minor_elinecolor')
        scolor(' Selected X-ray Line:', 'emph_elinecolor', newrow=True)
        scolor(' Held  X-ray Lines:',   'hold_elinecolor')
        scolor(' Pileup Prediction:',   'pileup_color', newrow=True)
        scolor(' Escape Prediction:',   'escape_color')

        panel.Add(HLine(panel, size=(400, 3)),  dcol=4, newrow=True)
        panel.Add(Button(panel, 'Done', size=(80, -1), action=self.onDone),
                  dcol=2, newrow=True)

        panel.pack()
        self.SetMinSize(panel.GetBestSize())
        self.Show()
        self.Raise()

    def onColor(self, event=None, item=None):
        color = hexcolor(event.GetValue())
        setattr(self.parent.conf, item, color)
        if item == 'spectra_color':
            self.parent.panel.conf.set_trace_color(color, trace=0)
        elif item == 'roi_color':
            self.parent.panel.conf.set_trace_color(color, trace=1)
        elif item == 'marker_color':
            for lmark in self.parent.cursor_markers:
                if lmark is not None:
                    lmark.set_color(color)

        elif item == 'roi_fillcolor' and self.parent.roi_patch is not None:
            self.parent.roi_patch.set_color(color)
        elif item == 'major_elinecolor':
            for l in self.parent.major_markers:
                l.set_color(color)
        elif item == 'minor_elinecolor':
            for l in self.parent.minor_markers:
                l.set_color(color)
        elif item == 'hold_elinecolor':
            for l in self.parent.hold_markers:
                l.set_color(color)

        self.parent.panel.canvas.draw()
        self.parent.panel.Refresh()

    def onDone(self, event=None):
        self.Destroy()

class XrayLinesFrame(wx.Frame):
    """settings frame for XRFDisplay"""

    k1lines = ['Ka1', 'Ka2', 'Kb1']
    k2lines = ['Kb2', 'Kb3']
    l1lines = ['La1', 'Lb1', 'Lb3', 'Lb4']
    l2lines = ['La2', 'Ll', 'Ln', 'Lb2,15']
    l3lines = ['Lg1', 'Lg2', 'Lg3']
    mlines  = ['Ma', 'Mb', 'Mg', 'Mz']

    def __init__(self, parent, size=(475, 325), **kws):
        self.parent = parent
        conf  = parent.conf
        kws['style'] = wx.DEFAULT_FRAME_STYLE
        wx.Frame.__init__(self, parent, -1, size=size,
                          title='XRF Line Selection', **kws)
        panel = GridPanel(self)
        self.checkbox = {}
        def add_elines(panel, lines, checked):
            for i in lines:
                cb = Check(panel, '%s ' % i, default = i in checked,
                           action=partial(self.onLine, label=i, lines=checked))
                self.checkbox[i] = cb
                panel.Add(cb, style=LEFT)

        hopts = {'size': (125, -1), 'bgcolour': (250, 250, 200)}
        labopts = {'newrow': True, 'style': LEFT}
        self.linedata = {'Major K Lines:': self.k1lines,
                         'Minor K Lines:': self.k2lines,
                         'Major L Lines:': self.l1lines,
                         'Minor L Lines:': self.l2lines+self.l3lines,
                         'Major M Lines:': self.mlines}
        panel.AddText(' Select X-ray Emission Lines', dcol=4, colour='#880000')
        panel.Add(HLine(panel, size=(450, 3)),  dcol=5, newrow=True)

        panel.Add(HyperText(panel, 'Major K Lines:',
                            action=partial(self.ToggleLines, lines=conf.K_major),
                            **hopts), **labopts)
        add_elines(panel, self.k1lines, conf.K_major)

        panel.Add(HyperText(panel, 'Minor K Lines:',
                            action=partial(self.ToggleLines, lines=conf.K_minor),
                            **hopts), **labopts)
        add_elines(panel, self.k2lines, conf.K_minor)

        panel.Add(HyperText(panel, 'Major L Lines:',
                            action=partial(self.ToggleLines, lines=conf.L_major),
                            **hopts), **labopts)
        add_elines(panel, self.l1lines, conf.L_major)

        panel.Add(HyperText(panel, 'Minor L Lines:',
                            action=partial(self.ToggleLines, lines=conf.L_minor),
                            **hopts), **labopts)
        add_elines(panel, self.l2lines, conf.L_minor)

        panel.AddText(' ', **labopts)
        add_elines(panel, self.l3lines, conf.L_minor)

        panel.Add(HyperText(panel, 'Major M Lines:',
                            action=partial(self.ToggleLines, lines=conf.M_major),
                            **hopts), **labopts)
        add_elines(panel, self.mlines,  conf.M_major)

        panel.AddText('Energy Range (keV): ', **labopts)
        fopts = {'minval':0, 'maxval':1000, 'precision':2, 'size':(75, -1)}
        panel.Add(FloatCtrl(panel, value=conf.e_min,
                            action=partial(self.onErange, is_max=False),
                            **fopts),  dcol=2, style=LEFT)

        panel.AddText(' : ')
        panel.Add(FloatCtrl(panel, value=conf.e_max,
                            action=partial(self.onErange, is_max=True),
                            **fopts), dcol=2, style=LEFT)

        panel.Add(HLine(panel, size=(450, 3)),  dcol=5, newrow=True)
        panel.Add(Button(panel, 'Done', size=(80, -1), action=self.onDone),
                  newrow=True, style=LEFT)
        panel.pack()
        self.Show()
        self.Raise()

    def ToggleLines(self, label=None, event=None, lines=None, **kws):
        if not event.leftIsDown:
            self.parent.Freeze()
            for line in self.linedata.get(label, []):
                check = not self.checkbox[line].IsChecked()
                self.checkbox[line].SetValue({True: 1, False:0}[check])
                self.onLine(checked=check, label=line, lines=lines)
            self.parent.Thaw()

    def onLine(self, event=None, checked=None, label=None, lines=None):
        if checked is None:
            try:
                checked =event.IsChecked()
            except:
                pass
        if checked is None: checked = False
        if lines is None:
            lines = []

        if label in lines and not checked:
            lines.remove(label)
        elif label not in lines and checked:
            lines.append(label)
        if self.parent.selected_elem is not None:
            self.parent.onShowLines(elem=self.parent.selected_elem)

    def onErange(self, event=None, value=None, is_max=True):
        if value is None:   return

        val = float(value)
        if is_max:
            self.parent.conf.e_max = val
        else:
            self.parent.conf.e_min = val
        if self.parent.selected_elem is not None:
            self.parent.onShowLines(elem=self.parent.selected_elem)


    def onDone(self, event=None):
        self.Destroy()

class XRFDisplayConfig:
    emph_elinecolor  = '#444444'
    major_elinecolor = '#DAD8CA'
    minor_elinecolor = '#F4DAC0'
    hold_elinecolor  = '#CAC8DA'
    marker_color     = '#77BB99'
    roi_fillcolor    = '#F8F0BA'
    roi_color        = '#d62728'
    spectra_color    = '#1f77b4'
    spectra2_color   = '#2ca02c'
    bgr_color        = '#ff7f0e'
    fit_color        = '#d62728'
    pileup_color     = '#555555'
    escape_color     = '#F07030'

    K_major = ['Ka1', 'Ka2', 'Kb1']
    K_minor = ['Kb3', 'Kb2']
    K_minor = []
    L_major = ['La1', 'Lb1', 'Lb3', 'Lb4']
    L_minor = ['Ln', 'Ll', 'Lb2,15', 'Lg2', 'Lg3', 'Lg1', 'La2']
    L_minor = []
    M_major = ['Ma', 'Mb', 'Mg', 'Mz']
    e_min   = 1.00
    e_max   = 30.0


class ROI_Averager():
    """ROI averager (over a fixed number of event samples)
       to give a rolling average of 'recent ROI values'

       roi_buff = ROI_Averager('13SDD1:mca1.R12',  nsamples=11)
       while True:
            print( roi_buff.average())

       typically, the ROIs update at a fixed 10 Hz, so 11 samples
       gives the ROI intensity integrated over the previous second

       using a ring buffer using numpy arrays
    """
    def __init__(self, nsamples=11):
        self.clear(nsamples = nsamples)

    def clear(self, nsamples=11):
        self.nsamples = nsamples
        self.index = -1
        self.lastval = 0
        self.toffset = time.time()
        self.data  =  np.zeros(self.nsamples, dtype=np.float32)
        self.times = -np.ones(self.nsamples, dtype=np.float32)

    def append(self, value):
        "adds value to ring buffer"
        idx = self.index = (self.index + 1) % self.data.size
        self.data[idx] = max(0, value - self.lastval)
        self.lastval  = value
        dt = time.time() - self.toffset
        # avoid first time point
        if (idx == 0 and max(self.times) < 0):
            dt = 0
        self.times[idx] =  dt

    def update(self, value):
        self.append(value)

    def get_mean(self):
        valid = np.where(self.times > 0)[0]
        return self.data[valid].mean()

    def get_cps(self):
        valid = np.where(self.times > 0)[0]
        if len(valid) < 1 or  self.times[valid].ptp() < 0.5:
            return 0
        return self.data[valid].sum() / self.times[valid].ptp()



#!/usr/bin/env python
"""
Enable wxPython to be used interacive by setting PyOS_InputHook.

based on inputhook.py and inputhookwx.py from IPython,
which has several authors, including
   Robin Dunn, Brian Granger, Ondrej Certik

tweaked by M Newville for larch
"""

import sys
import time
import signal
from select import select
from ctypes import c_void_p, c_int, cast, CFUNCTYPE, pythonapi

import wx
from wxutils import is_wxPhoenix

if is_wxPhoenix:
    is_wxMain = wx.IsMainThread
    wx_EvtLoop = wx.GUIEventLoop
else:
    is_wxMain = wx.Thread_IsMain
    wx_EvtLoop = wx.EventLoop

POLLTIME = 10 # milliseconds
ON_INTERRUPT = None
WXLARCH_SYM = None
WXLARCH_INP = None
UPDATE_GROUPNAME = '_sys.wx'
UPDATE_GROUP = None
UPDATE_VAR = 'force_wxupdate'
IN_MODAL_DIALOG = False

def update_requested():
    "check if update has been requested"
    global WXLARCH_SYM, UPDATE_VAR, UPDATE_GROUP, UPDATE_GROUPNAME
    if WXLARCH_SYM is not None:
        if UPDATE_GROUP is None:
            UPDATE_GROUP = WXLARCH_SYM.get_symbol(UPDATE_GROUPNAME)
        return getattr(UPDATE_GROUP, UPDATE_VAR, False)
    return False

def clear_update_request():
    "clear update request"
    global WXLARCH_SYM, UPDATE_VAR, UPDATE_GROUP, UPDATE_GROUPNAME
    if UPDATE_GROUP is not None:
        setattr(UPDATE_GROUP, UPDATE_VAR, False)

clock = time.time
sleep = time.sleep

def onCtrlC(*args, **kws):
    global WXLARCH_SYM
    try:
        WXLARCH_SYM.set_symbol('_sys.wx.keyboard_interrupt', True)
    except AttributeError:
        pass
    raise KeyboardInterrupt
    return 0

def capture_CtrlC():
    signal.signal(signal.SIGINT, onCtrlC)

def ignore_CtrlC():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

def allow_idle():
    # allow idle (needed for Mac OS X)
    pass

def stdin_ready():
    inp, out, err = select([sys.stdin],[],[],0)
    return bool(inp)

if sys.platform == 'win32':
    from msvcrt import kbhit as stdin_ready
    clock = time.monotonic
    def ignore_CtrlC():
        pass
elif sys.platform == 'darwin':
    from .allow_idle_macosx import allow_idle


class EnteredModalDialogHook(wx.ModalDialogHook):
    """
    set Global flag IN_MODAL_DIALOG when in a Modal Dialog.

    this will allow the event loop to *not* try to read stdina
    when a modal dialog is blocking, which causes problems on MacOS
    """
    def __init__(self):
        wx.ModalDialogHook.__init__(self)

    def Enter(self, dialog):
        global IN_MODAL_DIALOG
        IN_MODAL_DIALOG = True
        return wx.ID_NONE

    def Exit(self, dialog):
        global IN_MODAL_DIALOG
        IN_MODAL_DIALOG = False


class EventLoopRunner(object):
    def __init__(self, parent):
        self.parent = parent

    def run(self, poll_time=None):
        if poll_time is None:
            poll_time = POLLTIME
        self.t0 = clock()
        self.evtloop = wx_EvtLoop()
        self.timer = wx.Timer()
        self.parent.Bind(wx.EVT_TIMER, self.check_stdin)
        self.timer.Start(poll_time)
        self.evtloop.Run()

    def check_stdin(self, event=None):
        try:
            if (not IN_MODAL_DIALOG and (stdin_ready() or
                                        update_requested() or
                                        (clock() - self.t0) > 5)):
                self.timer.Stop()
                self.evtloop.Exit()
                del self.timer, self.evtloop
                clear_update_request()

        except KeyboardInterrupt:
            print('Captured Ctrl-C!')
        except:
            print(sys.exc_info())


def inputhook_wx():
    """Run the wx event loop by processing pending events only.

    This keeps processing pending events until stdin is ready.
    After processing all pending events, a call to time.sleep is inserted.
    This is needed, otherwise, CPU usage is at 100%.
    This sleep time should be tuned though for best performance.
    """
    # We need to protect against a user pressing Control-C when IPython is
    # idle and this is running. We trap KeyboardInterrupt and pass.
    try:
        app = wx.GetApp()
        if app is not None:
            assert is_wxMain()

            if not callable(signal.getsignal(signal.SIGINT)):
                signal.signal(signal.SIGINT, signal.default_int_handler)
            evtloop = wx_EvtLoop()
            ea = wx.EventLoopActivator(evtloop)
            t = clock()
            while not stdin_ready() and not update_requested():
                while evtloop.Pending():
                    t = clock()
                    evtloop.Dispatch()

                if callable(getattr(app, 'ProcessIdle', None)):
                    app.ProcessIdle()
                if callable(getattr(evtloop, 'ProcessIdle', None)):
                    evtloop.ProcessIdle()

                # We need to sleep at this point to keep the idle CPU load
                # low.  However, if sleep to long, GUI response is poor.
                used_time = clock() - t
                ptime = 0.001
                if used_time >  0.10: ptime = 0.05
                if used_time >  3.00: ptime = 0.25
                if used_time > 30.00: ptime = 1.00
                sleep(ptime)
            del ea
            clear_update_request()
    except KeyboardInterrupt:
        if callable(ON_INTERRUPT):
            ON_INTERRUPT()
    return 0

def inputhook_darwin():
    """Run the wx event loop, polling for stdin.

    This version runs the wx eventloop for an undetermined amount of time,
    during which it periodically checks to see if anything is ready on
    stdin.  If anything is ready on stdin, the event loop exits.

    The argument to eloop.run controls how often the event loop looks at stdin.
    This determines the responsiveness at the keyboard.  A setting of 20ms
    gives decent keyboard response.  It can be shortened if we know the loop will
    exit (as from update_requested()), but CPU usage of the idle process goes up
    if we run this too frequently.
    """
    global POLLTIME, ON_INTERRUPT
    try:
        app = wx.GetApp()
        if app is not None:
            assert is_wxMain()
            modal_hook = EnteredModalDialogHook()
            modal_hook.Register()
            eloop = EventLoopRunner(parent=app)
            ptime = POLLTIME
            if update_requested():
                ptime /= 10
            eloop.run(poll_time=ptime)
    except KeyboardInterrupt:
        print(" See KeyboardInterrupt from darwin hook")
        if callable(ON_INTERRUPT):
            ON_INTERRUPT()
    return 0

if sys.platform == 'darwin':
    # On OSX, evtloop.Pending() always returns True, regardless of there being
    # any events pending. As such we can't use implementations 1 or 3 of the
    # inputhook as those depend on a pending/dispatch loop.
    inputhook_wx = inputhook_darwin

try:
    capture_CtrlC()
except:
    pass
cback = CFUNCTYPE(c_int)(inputhook_wx)
py_inphook = c_void_p.in_dll(pythonapi, 'PyOS_InputHook')
py_inphook.value = cast(cback, c_void_p).value

# import for Darwin!
allow_idle()

def ping(timeout=0.001):
    "ping wx"
    try:
        t0 = clock()
        app = wx.GetApp()
        if app is not None:
            assert is_wxMain()
            # Make a temporary event loop and process system events until
            # there are no more waiting, then allow idle events (which
            # will also deal with pending or posted wx events.)
            evtloop = wx_EventLoop()
            ea = wx.EventLoopActivator(evtloop)
            t0 = clock()
            while clock()-t0 < timeout:
                evtloop.Dispatch()
            app.ProcessIdle()
            del ea
    except:
        pass



#!/usr/bin/env python
"""
merge groups, interpolating if necessary
"""
import os
import numpy as np
from larch import Group
from larch.math import interp, index_of

def merge_groups(grouplist, master=None, xarray='energy', yarray='mu',
                 kind='cubic', trim=True, calc_yerr=True, _larch=None):

    """merge arrays from a list of groups.

    Arguments
    ---------
     grouplist   list of groups to merge
     master      group to use for common x arrary [None -> 1st group]
     xarray      name of x-array for merge ['energy']
     yarray      name of y-array for merge ['mu']
     kind        interpolation kind ['cubic']
     trim        whether to trim to the shortest energy range [True]
     calc_yerr   whether to use the variance in the input as yerr [True]

    Returns
    --------
     group with x-array and y-array containing merged data.

    """
    if master is None:
        master = grouplist[0]

    xout = getattr(master, xarray)
    xmins = [min(xout)]
    xmaxs = [max(xout)]
    yvals = []

    for g in grouplist:
        x = getattr(g, xarray)
        y = getattr(g, yarray)
        yvals.append(interp(x, y, xout, kind=kind))
        xmins.append(min(x))
        xmaxs.append(max(x))

    yvals = np.array(yvals)
    yave = yvals.mean(axis=0)
    ystd = yvals.std(axis=0)

    if trim:
        xmin = min(xmins)
        xmax = min(xmaxs)
        ixmin = index_of(xout, xmin)
        ixmax = index_of(xout, xmax)
        xout = xout[ixmin:ixmax]
        yave = yave[ixmin:ixmax]
        ystd = ystd[ixmin:ixmax]

    grp = Group()
    setattr(grp, xarray, xout)
    setattr(grp, yarray, yave)
    setattr(grp, yarray + '_std', ystd)
    return grp



#!/usr/bin/env python
"""
Linear Combination panel
"""
import os
import time

import wx
import wx.lib.scrolledpanel as scrolled
import wx.dataview as dv
import numpy as np

from functools import partial
from collections import OrderedDict

import lmfit
from lmfit.printfuncs import gformat, fit_report

from larch import Group
from larch.math import index_of

from larch.wxlib import (BitmapButton, FloatCtrl, FloatSpin, ToggleButton,
                         GridPanel, get_icon, SimpleText, pack, Button, HLine,
                         Choice, Check, CEN, LEFT, Font, FONTSIZE,
                         MenuItem, FRAMESTYLE, GUIColors, FileSave,
                         EditableListBox)

from .taskpanel import TaskPanel, DataTableGrid
from larch.io.columnfile import write_ascii

np.seterr(all='ignore')

# plot options:
norm   = 'Normalized \u03bC(E)'
rawmu  = 'Raw \u03bC(E)'
dmude  = 'd\u03bC(E)/dE'
chik   = '\u03c7(k)'
noplot = '<no plot>'
noname = '<none>'

FitSpace_Choices = [norm, rawmu, dmude, chik]
Plot_Choices = ['Data + Sum', 'Data + Sum + Components']

DVSTYLE = dv.DV_SINGLE|dv.DV_VERT_RULES|dv.DV_ROW_LINES

defaults = dict(elo=-5.e5, ehi=5.e5, elo_rel=-40, ehi_rel=110,
                fitspace=norm, all_combos=True, sum_to_one=True,
                show_e0=False, show_fitrange=True)

MAX_COMPONENTS = 20

def make_lcfplot(dgroup, form, nfit=0):
    """make larch plot commands to plot LCF fit from form"""
    form['group'] = dgroup.groupname
    form['filename'] = dgroup.filename

    form['nfit'] = nfit
    form['label'] = label = 'Fit #%2.2d' % (nfit+1)
    form['plotopt'] = 'show_norm=True'
    if form['arrayname'] == 'dmude':
        form['plotopt'] = 'show_deriv=True'

    erange = form['ehi'] - form['elo']
    form['pemin'] = 10*int( (form['elo'] - 5 - erange/4.0) / 10.0)
    form['pemax'] = 10*int( (form['ehi'] + 5 + erange/4.0) / 10.0)


    cmds = ["""plot_mu({group:s}, {plotopt:s}, delay_draw=True, label='data',
    emin={pemin:.1f}, emax={pemax:.1f}, title='{filename:s}, {label:s}')"""]

    if hasattr(dgroup, 'lcf_result'):
        with_comps = True # "Components" in form['plotchoice']
        delay = 'delay_draw=True' if with_comps else 'delay_draw=False'
        xarr = "{group:s}.lcf_result[{nfit:d}].xdata"
        yfit = "{group:s}.lcf_result[{nfit:d}].yfit"
        ycmp = "{group:s}.lcf_result[{nfit:d}].ycomps"
        cmds.append("plot(%s, %s, label='%s', zorder=30, %s)" % (xarr, yfit, label, delay))
        ncomps = len(dgroup.lcf_result[nfit].ycomps)
        if with_comps:
            for i, key in enumerate(dgroup.lcf_result[nfit].ycomps):
                delay = 'delay_draw=False' if i==(ncomps-1) else 'delay_draw=True'
                cmds.append("plot(%s, %s['%s'], label='%s', %s)" % (xarr, ycmp, key, key, delay))

    # if form['show_e0']:
    #     cmds.append("plot_axvline({e0:1f}, color='#DDDDCC', zorder=-10)")
    if form['show_fitrange']:
        cmds.append("plot_axvline({elo:1f}, color='#EECCCC', zorder=-10)")
        cmds.append("plot_axvline({ehi:1f}, color='#EECCCC', zorder=-10)")

    script = "\n".join(cmds)
    return script.format(**form)

class ResultFrame(wx.Frame):
    def __init__(self, parent=None,  **kws):

        wx.Frame.__init__(self, None, -1, title='Linear Combination Results',
                          style=FRAMESTYLE, size=(825, 675), **kws)
        self.parent = parent
        self.datasets = {}
        self.datagroup = None
        self.form = None
        self.larch_eval = None
        self.current_fit = 0
        self.createMenus()
        self.build()

    def createMenus(self):
        self.menubar = wx.MenuBar()
        fmenu = wx.Menu()
        m = {}

        MenuItem(self, fmenu, "Save Fit And Components for Current Group",
                 "Save Fit and Compoents to Data File for Current Group",  self.onSaveGroupFit)

        MenuItem(self, fmenu, "Save Statistics for Best N Fits for Current Group",
                 "Save Statistics and Weights for Best N Fits for Current Group",  self.onSaveGroupStats)

        MenuItem(self, fmenu, "Save Data and Best N Fits for Current Group",
                 "Save Data and Best N Fits for Current Group",  self.onSaveGroupMultiFits)

        fmenu.AppendSeparator()
        MenuItem(self, fmenu, "Save Statistics Report for All Fitted Groups",
                 "Save Statistics for All Fitted Groups",  self.onSaveAllStats)

        self.menubar.Append(fmenu, "&File")
        self.SetMenuBar(self.menubar)

    def build(self):
        sizer = wx.GridBagSizer(5, 5)
        sizer.SetVGap(5)
        sizer.SetHGap(5)

        splitter = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(200)

        self.datalistbox = EditableListBox(splitter, self.ShowDataSet,
                                           size=(250, -1))
        panel = scrolled.ScrolledPanel(splitter)


        self.SetMinSize((650, 600))
        self.colors = GUIColors()

        # title row
        self.wids = wids = {}
        title = SimpleText(panel, 'Linear Combination Results',
                           font=Font(FONTSIZE+2),
                           colour=self.colors.title, style=LEFT)

        wids['plot_one'] = Button(panel, 'Plot This Fit', size=(125, -1),
                                  action=self.onPlotOne)
        wids['plot_sel'] = Button(panel, 'Plot N Best Fits', size=(125, -1),
                                  action=self.onPlotSel)

        wids['plot_ntitle'] = SimpleText(panel, 'N fits to plot: ')

        wids['plot_nchoice'] = Choice(panel, size=(60, -1),
                                      choices=['%d' % i for i in range(1, 21)])
        wids['plot_nchoice'].SetStringSelection('5')

        wids['data_title'] = SimpleText(panel, '<--> ',  font=Font(FONTSIZE+2),
                                        colour=self.colors.title, style=LEFT)
        wids['nfits_title'] = SimpleText(panel, 'showing 5 best fits')

        copts = dict(size=(125, 30), default=True, action=self.onPlotOne)
        # wids['show_e0'] = Check(panel, label='show E0?', **copts)
        wids['show_fitrange'] = Check(panel, label='show fit range?', **copts)

        irow = 0
        sizer.Add(title,              (irow, 0), (1, 1), LEFT)
        sizer.Add(wids['data_title'], (irow, 1), (1, 2), LEFT)

        irow += 1
        sizer.Add(wids['nfits_title'],     (irow, 0), (1, 1), LEFT)


        irow += 1
        self.wids['paramstitle'] = SimpleText(panel, '[[Parameters]]',
                                              font=Font(FONTSIZE+2),
                                              colour=self.colors.title, style=LEFT)
        sizer.Add(self.wids['paramstitle'], (irow, 0), (1, 1), LEFT)


        pview = self.wids['params'] = dv.DataViewListCtrl(panel, style=DVSTYLE)
        pview.SetMinSize((475, 200))
        pview.AppendTextColumn(' Parameter ', width=230)
        pview.AppendTextColumn(' Best-Fit Value', width=120)
        pview.AppendTextColumn(' Standard Error ', width=120)
        for col in range(3):
            this = pview.Columns[col]
            isort, align = True, wx.ALIGN_RIGHT
            if col == 0:
                align = wx.ALIGN_LEFT
            this.Sortable = isort
            this.Alignment = this.Renderer.Alignment = align


        irow += 1
        sizer.Add(self.wids['params'],       (irow,   0), (4, 1), LEFT)
        sizer.Add(self.wids['plot_one'],     (irow,   1), (1, 2), LEFT)
        sizer.Add(self.wids['plot_sel'],     (irow+1, 1), (1, 2), LEFT)
        sizer.Add(self.wids['plot_ntitle'],  (irow+2, 1), (1, 1), LEFT)
        sizer.Add(self.wids['plot_nchoice'], (irow+2, 2), (1, 1), LEFT)
        # sizer.Add(self.wids['show_e0'],      (irow+3, 1), (1, 2), LEFT)
        sizer.Add(self.wids['show_fitrange'],(irow+3, 1), (1, 2), LEFT)

        irow += 4
        sizer.Add(HLine(panel, size=(675, 3)), (irow, 0), (1, 4), LEFT)

        sview = self.wids['stats'] = dv.DataViewListCtrl(panel, style=DVSTYLE)
        sview.Bind(dv.EVT_DATAVIEW_SELECTION_CHANGED, self.onSelectFitStat)
        sview.AppendTextColumn(' Fit #', width=50)
        sview.AppendTextColumn(' N_vary', width=50)
        sview.AppendTextColumn(' N_eval', width=60)
        sview.AppendTextColumn(' \u03c7\u00B2', width=110)
        sview.AppendTextColumn(' \u03c7\u00B2_reduced', width=110)
        sview.AppendTextColumn(' Akaike Info', width=110)
        sview.AppendTextColumn(' Bayesian Info', width=110)

        for col in range(sview.ColumnCount):
            this = sview.Columns[col]
            isort, align = True, wx.ALIGN_RIGHT
            if col == 0:
                align = wx.ALIGN_CENTER
            this.Sortable = isort
            this.Alignment = this.Renderer.Alignment = align

        sview.SetMinSize((675, 175))

        irow += 1
        title = SimpleText(panel, '[[Fit Statistics]]', font=Font(FONTSIZE+2),
                           colour=self.colors.title, style=LEFT)
        sizer.Add(title, (irow, 0), (1, 4), LEFT)

        irow += 1
        sizer.Add(sview, (irow, 0), (1, 4), LEFT)

        irow += 1
        sizer.Add(HLine(panel, size=(675, 3)), (irow, 0), (1, 4), LEFT)

        irow += 1
        title = SimpleText(panel, '[[Weights]]', font=Font(FONTSIZE+2),
                           colour=self.colors.title, style=LEFT)
        sizer.Add(title, (irow, 0), (1, 4), LEFT)
        self.wids['weightspanel'] = ppan = wx.Panel(panel)

        p1 = SimpleText(ppan, ' < Weights > ')
        os = wx.BoxSizer(wx.VERTICAL)
        os.Add(p1, 1, 3)
        pack(ppan, os)
        ppan.SetMinSize((675, 175))

        irow += 1
        sizer.Add(ppan, (irow, 0), (1, 4), LEFT)

        irow += 1
        sizer.Add(HLine(panel, size=(675, 3)), (irow, 0), (1, 4), LEFT)

        pack(panel, sizer)
        panel.SetupScrolling()

        splitter.SplitVertically(self.datalistbox, panel, 1)

        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(splitter, 1, wx.GROW|wx.ALL, 5)

        pack(self, mainsizer)
        # self.SetSize((725, 750))
        self.Show()
        self.Raise()

    def ShowDataSet(self, evt=None):
        dataset = evt.GetString()
        group = self.datasets.get(evt.GetString(), None)
        if group is not None:
            self.show_results(datagroup=group)

    def add_results(self, dgroup, form=None, larch_eval=None, show=True):
        name = dgroup.filename
        if name not in self.datalistbox.GetItems():
            self.datalistbox.Append(name)
        self.datasets[name] = dgroup
        if show:
            self.show_results(datagroup=dgroup, form=form, larch_eval=larch_eval)

    def show_results(self, datagroup=None, form=None, larch_eval=None):
        if datagroup is not None:
            self.datagroup = datagroup
        if form is not None:
            self.form = form
        if larch_eval is not None:
            self.larch_eval = larch_eval

        form = self.form
        datagroup = self.datagroup
        lcf_history = getattr(self.datagroup, 'lcf_history', [])

        wids = self.wids
        wids['data_title'].SetLabel(self.datagroup.filename)
        # wids['show_e0'].SetValue(form['show_e0'])
        wids['show_fitrange'].SetValue(form['show_fitrange'])

        wids['stats'].DeleteAllItems()
        results = self.datagroup.lcf_result[:20]
        self.nresults = len(results)
        wids['nfits_title'].SetLabel('showing %i best results' % self.nresults)

        for i, res in enumerate(results):
            args = ['%2.2d' % (i+1)]
            for attr in ('nvarys', 'nfev', 'chisqr', 'redchi', 'aic', 'bic'):
                val = getattr(res.result, attr)
                if isinstance(val, int):
                    val = '%d' % val
                else:
                    val = gformat(val, 12)
                args.append(val)
            wids['stats'].AppendItem(tuple(args))

        wpan = self.wids['weightspanel']
        wpan.DestroyChildren()

        wview = self.wids['weights'] = dv.DataViewListCtrl(wpan, style=DVSTYLE)
        wview.Bind(dv.EVT_DATAVIEW_SELECTION_CHANGED, self.onSelectFitParam)
        wview.AppendTextColumn(' Fit #', width=50)

        for i, cname in enumerate(form['comp_names']):
            wview.AppendTextColumn(cname, width=100)
        wview.AppendTextColumn('Total', width=100)

        for col in range(len(form['comp_names'])+2):
            this = wview.Columns[col]
            isort, align = True, wx.ALIGN_RIGHT
            if col == 0:
                align = wx.ALIGN_CENTER
            this.Sortable = isort
            this.Alignment = this.Renderer.Alignment = align

        for i, res in enumerate(results):
            args = ['%2.2d' % (i+1)]
            for cname in form['comp_names'] + ['total']:
                val = '--'
                if cname in res.params:
                    val = "%.4f" % res.params[cname].value
                args.append(val)
            wview.AppendItem(tuple(args))

        os = wx.BoxSizer(wx.VERTICAL)
        os.Add(wview, 1, wx.GROW|wx.ALL)
        pack(wpan, os)

        wview.SetMinSize((675, 200))
        s1, s2 = self.GetSize()
        if s2 % 2 == 0:
            s2 = s2 + 1
        else:
            s2 = s2 - 1
        self.SetSize((s1, s2))
        self.show_fitresult(0)
        self.Refresh()

    def onSelectFitParam(self, evt=None):
        if self.wids['weights'] is None:
            return
        item = self.wids['weights'].GetSelectedRow()
        self.show_fitresult(item)

    def onSelectFitStat(self, evt=None):
        if self.wids['stats'] is None:
            return
        item = self.wids['stats'].GetSelectedRow()
        self.show_fitresult(item)

    def show_fitresult(self, n):
        fit_result = self.datagroup.lcf_result[n]
        self.current_fit = n
        wids = self.wids
        wids['nfits_title'].SetLabel('Showing Fit # %2.2d' % (n+1))
        wids['paramstitle'].SetLabel('[[Parameters for Fit # %2.2d]]' % (n+1))

        wids['params'].DeleteAllItems()

        for pname, par in fit_result.params.items():
            args = [pname, gformat(par.value, 12), '--']
            if par.stderr is not None:
                args[2] = gformat(par.stderr, 12)
            self.wids['params'].AppendItem(tuple(args))

    def onPlotOne(self, evt=None):
        if self.form is None or self.larch_eval is None:
            return

        self.form['show_fitrange'] = self.wids['show_fitrange'].GetValue()
        self.larch_eval(make_lcfplot(self.datagroup,
                                     self.form, nfit=self.current_fit))

    def onPlotSel(self, evt=None):
        if self.form is None or self.larch_eval is None:
            return
        self.form['show_fitrange'] = self.wids['show_fitrange'].GetValue()
        form = self.form
        dgroup = self.datagroup

        form['plotopt'] = 'show_norm=True'
        if form['arrayname'] == 'dmude':
            form['plotopt'] = 'show_deriv=True'

        erange = form['ehi'] - form['elo']
        form['pemin'] = 10*int( (form['elo'] - 5 - erange/4.0) / 10.0)
        form['pemax'] = 10*int( (form['ehi'] + 5 + erange/4.0) / 10.0)

        cmds = ["""plot_mu({group:s}, {plotopt:s}, delay_draw=True, label='data',
        emin={pemin:.1f}, emax={pemax:.1f}, title='{filename:s}')"""]

        nfits = int(self.wids['plot_nchoice'].GetStringSelection())
        for i in range(nfits):
            delay = 'delay_draw=True' if i<nfits-1 else 'delay_draw=False'
            xarr = "{group:s}.lcf_result[%i].xdata" % i
            yfit = "{group:s}.lcf_result[%i].yfit" % i
            lab = 'Fit #%2.2d' % (i+1)
            cmds.append("plot(%s, %s, label='%s', zorder=30, %s)" % (xarr, yfit, lab, delay))

        # if form['show_e0']:
        #     cmds.append("plot_axvline({e0:1f}, color='#DDDDCC', zorder=-10)")
        if form['show_fitrange']:
            cmds.append("plot_axvline({elo:1f}, color='#EECCCC', zorder=-10)")
            cmds.append("plot_axvline({ehi:1f}, color='#EECCCC', zorder=-10)")

        script = "\n".join(cmds)
        self.larch_eval(script.format(**form))

    def onSaveGroupFit(self, evt=None):
        "Save Fit and Compoents for current fit to Data File"
        nfit = self.current_fit
        dgroup = self.datagroup
        nfits = int(self.wids['plot_nchoice'].GetStringSelection())

        deffile = "%s_LinearFit%i.dat" % (dgroup.filename, nfit+1)
        wcards  = 'Data Files (*.dat)|*.dat|All files (*.*)|*.*'
        path = FileSave(self, 'Save Fit and Components to File',
                        default_file=deffile, wildcard=wcards)
        if path is None:
            return

        form  = self.form
        label = [' energy         ',
                 ' data           ',
                 ' best_fit       ']
        result = dgroup.lcf_result[nfit]

        header = ['Larch Linear Fit Result for Fit: #%2.2d' % (nfit+1),
                  'Dataset filename: %s ' % dgroup.filename,
                  'Larch group: %s ' % dgroup.groupname,
                  'Array name: %s' %  form['arrayname'],
                  'Energy fit range: [%f, %f]' % (form['elo'], form['ehi']),
                  'Components: ']
        for key, val in result.weights.items():
            header.append('  %s: %f' % (key, val))

        report = fit_report(result.result).split('\n')
        header.extend(report)

        out = [result.xdata, result.ydata, result.yfit]
        for compname, compdata in result.ycomps.items():
            label.append(' %s' % (compname + ' '*(max(1, 15-len(compname)))))
            out.append(compdata)

        label = ' '.join(label)
        _larch = self.parent.controller.larch
        write_ascii(path, header=header, label=label, _larch=_larch, *out)


    def onSaveGroupStats(self, evt=None):
        "Save Statistics and Weights for Best N Fits for the current group"
        dgroup = self.datagroup
        nfits = int(self.wids['plot_nchoice'].GetStringSelection())
        results = dgroup.lcf_result[:nfits]
        nresults = len(results)
        deffile = "%s_LinearStats%i.dat" % (dgroup.filename, nresults)
        wcards  = 'Data Files (*.dat)|*.dat|All files (*.*)|*.*'

        path = FileSave(self, 'Save Statistics and Weights for Best N Fits',
                        default_file=deffile, wildcard=wcards)
        if path is None:
            return
        form = self.form

        header = ['Larch Linear Fit Statistics for %2.2d best results' % (nresults),
                  'Dataset filename: %s ' % dgroup.filename,
                  'Larch group: %s ' % dgroup.groupname,
                  'Array name: %s' %  form['arrayname'],
                  'Energy fit range: [%f, %f]' % (form['elo'], form['ehi']),
                  'N_Data: %d' % len(results[0].xdata)]

        label = ['fit #', 'n_varys', 'n_eval', 'chi2',
                  'chi2_reduced', 'akaike_info', 'bayesian_info']
        label.extend(form['comp_names'])
        label.append('Total')
        for i in range(len(label)):
            if len(label[i]) < 13:
                label[i] = (" %s                " % label[i])[:13]
        label = ' '.join(label)

        out = []
        for i, res in enumerate(results):
            dat = [(i+1)]
            for attr in ('nvarys', 'nfev', 'chisqr', 'redchi', 'aic', 'bic'):
                dat.append(getattr(res.result, attr))
            for cname in form['comp_names'] + ['total']:
                val = 0.0
                if cname in res.params:
                    val = res.params[cname].value
                dat.append(val)
            out.append(dat)

        out = np.array(out).transpose()
        _larch = self.parent.controller.larch
        write_ascii(path, header=header, label=label, _larch=_larch, *out)

    def onSaveGroupMultiFits(self, evt=None):
        "Save Data and Best N Fits for the current group"
        dgroup = self.datagroup
        nfits = int(self.wids['plot_nchoice'].GetStringSelection())
        results = dgroup.lcf_result[:nfits]
        nresults = len(results)

        deffile = "%s_LinearFits%i.dat" % (dgroup.filename, nresults)
        wcards  = 'Data Files (*.dat)|*.dat|All files (*.*)|*.*'

        path = FileSave(self, 'Save Best N Fits',
                        default_file=deffile, wildcard=wcards)
        if path is None:
            return
        form = self.form
        header = ['Larch Linear Arrays for %2.2d best results' % (nresults),
                  'Dataset filename: %s ' % dgroup.filename,
                  'Larch group: %s ' % dgroup.groupname,
                  'Array name: %s' %  form['arrayname'],
                  'Energy fit range: [%f, %f]' % (form['elo'], form['ehi'])]

        label = [' energy         ',  ' data           ']
        label.extend([' fit_%2.2d         ' % i for i in range(nresults)])
        label = ' '.join(label)

        out = [results[0].xdata, results[0].ydata]
        for i, res in enumerate(results):
            out.append(results[i].yfit)

        _larch = self.parent.controller.larch
        write_ascii(path, header=header, label=label, _larch=_larch, *out)

    def onSaveAllStats(self, evt=None):
        "Save All Statistics and Weights "
        deffile = "LinearFitStats.csv"
        wcards  = 'CVS Files (*.csv)|*.csv|All files (*.*)|*.*'
        path = FileSave(self, 'Save Statistics Report',
                        default_file=deffile, wildcard=wcards)
        if path is None:
            return
        form = self.form

        out = ['# Larch Linear Fit Statistics Report (best results) %s' % time.ctime(),
               '# Array name: %s' %  form['arrayname'],
               '# Energy fit range: [%f, %f]' % (form['elo'], form['ehi'])]

        label = [('Data Set' + ' '*25)[:25],
                 'n_varys', 'chi-square',
                 'chi-square_red', 'akaike_info', 'bayesian_info']
        label.extend(form['comp_names'])
        label.append('Total')
        for i in range(len(label)):
            if len(label[i]) < 12:
                label[i] = (" %s                " % label[i])[:12]
        label = ', '.join(label)
        out.append('# %s' % label)

        for name, dgroup in self.datasets.items():
            res = dgroup.lcf_result[0]
            label = dgroup.filename
            if len(label) < 25:
                label = (label + ' '*25)[:25]
            dat = [label]
            for attr in ('nvarys', 'chisqr', 'redchi', 'aic', 'bic'):
                dat.append(gformat(getattr(res.result, attr), 12))
            for cname in form['comp_names'] + ['total']:
                val = 0
                if cname in res.params:
                    val = res.params[cname].value
                dat.append(gformat(val, 12))
            out.append(', '.join(dat))
        out.append('')

        with open(path, 'w') as fh:
            fh.write('\n'.join(out))

class LinearComboPanel(TaskPanel):
    """Liear Combination Panel"""
    def __init__(self, parent, controller, **kws):
        TaskPanel.__init__(self, parent, controller,
                           configname='lincombo_config',
                           config=defaults, **kws)

    def process(self, dgroup, **kws):
        """ handle linear combo processing"""
        if self.skip_process:
            return
        return self.read_form()

    def build_display(self):
        panel = self.panel
        wids = self.wids
        self.skip_process = True

        wids['fitspace'] = Choice(panel, choices=FitSpace_Choices,
                                  size=(175, -1))
        wids['fitspace'].SetStringSelection(norm)

        add_text = self.add_text

        opts = dict(digits=2, increment=1.0, relative_e0=False) # True)

        elo_wids = self.add_floatspin('elo', value=defaults['elo'], **opts)
        ehi_wids = self.add_floatspin('ehi', value=defaults['ehi'], **opts)

        wids['fit_group'] = Button(panel, 'Fit this Group', size=(175, -1),
                                   action=self.onFitOne)
        wids['fit_selected'] = Button(panel, 'Fit Selected Groups', size=(175, -1),
                                      action=self.onFitAll)

        wids['add_selected'] = Button(panel, 'Use Selected Groups as Components',
                                      size=(300, -1),
                                      action=self.onUseSelected)

        wids['saveconf'] = Button(panel, 'Save as Default Settings', size=(225, -1),
                                  action=self.onSaveConfigBtn)

        opts = dict(default=True, size=(75, -1), action=self.onPlotOne)

        wids['show_fitrange'] = Check(panel, label='show?', **opts)

        wids['sum_to_one'] = Check(panel, label='Weights Must Sum to 1?', default=True)
        wids['all_combos'] = Check(panel, label='Fit All Combinations?', default=True)
        max_ncomps = self.add_floatspin('max_ncomps', value=10, digits=0, increment=1,
                                        min_val=0, max_val=20, size=(60, -1),
                                        with_pin=False)

        panel.Add(SimpleText(panel, ' Linear Combination Analysis',
                             **self.titleopts), dcol=4)

        add_text('Array to Fit: ', newrow=True)
        panel.Add(wids['fitspace'], dcol=3)

        add_text('Fit Energy Range: ')
        panel.Add(elo_wids)
        add_text(' : ', newrow=False)
        panel.Add(ehi_wids)
        panel.Add(wids['show_fitrange'])

        panel.Add(HLine(panel, size=(625, 3)), dcol=5, newrow=True)
        add_text('Run Fit: ')
        panel.Add(wids['fit_group'], dcol=2)
        panel.Add(wids['fit_selected'], dcol=3)
        add_text('Fit Options: ')
        panel.Add(wids['sum_to_one'], dcol=2)
        panel.Add((10, 10), dcol=1, newrow=True)
        panel.Add(wids['all_combos'], dcol=2)
        add_text('Max # Components: ', newrow=False)
        panel.Add(max_ncomps, dcol=2)

        panel.Add(HLine(panel, size=(625, 3)), dcol=5, newrow=True)

        add_text('Components: ')
        panel.Add(wids['add_selected'], dcol=4)

        collabels = [' File /Group Name   ', 'weight', 'min', 'max']
        colsizes = [300, 80, 80, 80]
        coltypes = ['str', 'float:12,4', 'float:12,4', 'float:12,4']
        coldefs  = ['', 1.0/MAX_COMPONENTS, 0.0, 1.0]

        wids['table'] = DataTableGrid(panel, nrows=MAX_COMPONENTS,
                                      collabels=collabels,
                                      datatypes=coltypes, defaults=coldefs,
                                      colsizes=colsizes)

        wids['table'].SetMinSize((625, 250))
        panel.Add(wids['table'], newrow=True, dcol=6)

        panel.Add(HLine(panel, size=(625, 3)), dcol=5, newrow=True)
        panel.Add(wids['saveconf'], dcol=4, newrow=True)
        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((10, 10), 0, LEFT, 3)
        sizer.Add(panel, 1, LEFT, 3)
        pack(self, sizer)
        self.skip_process = False

    def onComponent(self, evt=None, comp=None):
        if comp is None or evt is None:
            return

        comps = []
        for wname, wid in self.wids.items():
            if wname.startswith('compchoice'):
                pref, n = wname.split('_')
                if wid.GetSelection() > 0:
                    comps.append((int(n), wid.GetStringSelection()))
                else:
                    self.wids["compval_%s" % n].SetValue(0)

        cnames = set([elem[1] for elem in comps])
        if len(cnames) < len(comps):
            comps.remove((comp, evt.GetString()))
            self.wids["compchoice_%2.2d" % comp].SetSelection(0)

        weight = 1.0 / len(comps)

        for n, cname in comps:
            self.wids["compval_%2.2d" % n].SetValue(weight)


    def fill_form(self, dgroup):
        """fill in form from a data group"""
        opts = self.get_config(dgroup)
        self.dgroup = dgroup
        if isinstance(dgroup, Group):
            d_emin = min(dgroup.energy)
            d_emax = max(dgroup.energy)
            if opts['elo'] < d_emin:
                opts['elo'] = defaults['elo_rel'] + int(dgroup.e0/10.0)*10
            if opts['ehi'] > d_emax:
                opts['ehi'] =  defaults['ehi_rel'] + int(dgroup.e0/10.0)*10

        self.skip_process = True
        wids = self.wids
        for attr in ('elo', 'ehi'):
            val = opts.get(attr, None)
            if val is not None:
                wids[attr].SetValue(val)

        for attr in ('all_combos', 'sum_to_one', 'show_fitrange'):
            wids[attr].SetValue(opts.get(attr, True))

        for attr in ('fitspace',):
            if attr in opts:
                wids[attr].SetStringSelection(opts[attr])

#         groupnames = [noname] + list(self.controller.file_groups.keys())
#         for wname, wid in wids.items():
#             if wname.startswith('compchoice'):
#                 cur = wid.GetStringSelection()
#                 wid.Clear()
#                 wid.AppendItems(groupnames)
#                 if cur in groupnames:
#                     wid.SetStringSelection(cur)
#                 else:
#                     wid.SetSelection(0)
#             elif wname.startswith('comp'):
#                 wid.SetValue(opts.get(wname, wid.GetValue()))
        self.skip_process = False

    def read_form(self, dgroup=None):
        "read form, return dict of values"
        self.skip_process = True
        if dgroup is None:
            dgroup = self.controller.get_group()
        self.dgroup = dgroup
        opts = {'group': dgroup.groupname, 'filename':dgroup.filename}
        wids = self.wids
        for attr in ('elo', 'ehi', 'max_ncomps'):
            opts[attr] = wids[attr].GetValue()

        for attr in ('fitspace', ): # 'plotchoice'):
            opts[attr] = wids[attr].GetStringSelection()
        for attr in ('all_combos', 'sum_to_one', 'show_fitrange'):
            opts[attr] = wids[attr].GetValue()

        for attr, wid in wids.items():
            if attr.startswith('compchoice'):
                opts[attr] = wid.GetStringSelection()
            elif attr.startswith('comp'):
                opts[attr] = wid.GetValue()

        comps, cnames, wval, wmin, wmax = [], [], [], [], []

        table_data = self.wids['table'].table.data
        for _cname, _wval, _wmin, _wmax in table_data:
            if _cname.strip() in ('', None) or len(_cname) < 1:
                break
            cnames.append(_cname)
            comps.append(self.controller.file_groups[_cname])
            wval.append("%.5f" % _wval)
            wmin.append("%.5f" % _wmin)
            wmax.append("%.5f" % _wmax)

        opts['comp_names'] = cnames
        opts['comps']   = ', '.join(comps)
        opts['weights'] = ', '.join(wval)
        opts['minvals'] = ', '.join(wmin)
        opts['maxvals'] = ', '.join(wmax)
        opts['func'] = 'lincombo_fit'
        if opts['all_combos']:
            opts['func'] = 'lincombo_fitall'

        opts['arrayname'] = 'norm'
        if opts['fitspace'] == dmude:
            opts['arrayname'] = 'dmude'
        elif opts['fitspace'] == rawmu:
            opts['arrayname'] = 'mu'
        elif opts['fitspace'] == chik:
            opts['arrayname'] = 'chi'

        self.skip_process = False
        return opts

    def onSaveConfigBtn(self, evt=None):
        conf = self.get_config()
        conf.update(self.read_form())
        self.set_defaultconfig(conf)

    def onUseSelected(self, event=None):
        """ use selected groups as standards"""
        self.skip_process = True
        selected_groups = self.controller.filelist.GetCheckedStrings()
        if len(selected_groups) == 0:
            return
        if len(selected_groups) >= MAX_COMPONENTS:
            selected_groups = selected_groups[:MAX_COMPONENTS]
        weight = 1.0/len(selected_groups)

        grid_data = []
        for grp in selected_groups:
            grid_data.append([grp, weight, 0, 1])

        self.wids['table'].table.data = grid_data
        self.wids['table'].table.View.Refresh()
        self.skip_process = False


    def do_fit(self, groupname, form):
        """run lincombo fit for a group"""
        form['gname'] = groupname
        script = """# do LCF for {gname:s}
result = {func:s}({gname:s}, [{comps:s}],
            xmin={elo:.4f}, xmax={ehi:.4f},
            arrayname='{arrayname:s}',
            sum_to_one={sum_to_one},
            weights=[{weights:s}],
            minvals=[{minvals:s}],
            maxvals=[{maxvals:s}],
            max_ncomps={max_ncomps:.0f})
"""
        if form['all_combos']:
            script = "%s\n{gname:s}.lcf_result = result\n" % script
        else:
            script = "%s\n{gname:s}.lcf_result = [result]\n" % script

        self.larch_eval(script.format(**form))

        dgroup = self.controller.get_group(groupname)
        self.show_subframe('lcf_result',  ResultFrame)

        self.subframes['lcf_result'].add_results(dgroup, form=form,
                                                 larch_eval=self.larch_eval)
        self.plot(dgroup=dgroup)

    def onFitOne(self, event=None):
        """ handle process events"""
        if self.skip_process:
            return

        self.skip_process = True
        form = self.read_form()
        self.do_fit(form['group'], form)
        self.skip_process = False

    def onFitAll(self, event=None):
        """ handle process events"""
        if self.skip_process:
            return
        self.skip_process = True
        form = self.read_form()
        for sel in self.controller.filelist.GetCheckedStrings():
            gname = self.controller.file_groups[sel]
            self.do_fit(gname, form)
        self.skip_process = False

    def plot(self, dgroup=None):
        if self.skip_plotting:
            return
        self.onPlot(dgroup=dgroup)

    def onPlot(self, evt=None, dgroup=None):
        if self.skip_plotting:
            return
        if dgroup is None:
            dgroup = self.controller.get_group()
        form = self.read_form(dgroup=dgroup)
        script = make_lcfplot(dgroup, form, nfit=0)
        self.larch_eval(script)



#!/usr/bin/env python
#
# Cauchy Wavelet for EXAFS, adopted from
# matlab code from Munoz, Argoul, and Farges:
#
# CONTINUOUS CAUCHY WAVELET TRANSFORM OF EXAFS SIGNAL
# code freely downloaded from http://www.univ-mlv.fr/~farges/waw
# (c) 2000, Univ. Marne la Vallee, France
#
#  please cite us of this code with:
#   Munoz M., Argoul P. and Farges F.
#   Continuous Cauchy wavelet transform analyses of
#   EXAFS spectra: a qualitative approach.
#   American Mineralogist 88, pp. 694-700 (2003).
#
# version history:
# 1999 Hans-Argoul : core wavelet algorithm
# 1999-2002 Argoul-Munoz : EXAFS adapation
# 2002 Farges : graphical and user interface
# 2003 Munoz : CPU optimizations and graphical updates
# 2003 Farges-Munoz : various fixes and web version
#
# 2014-Apr M Newville : translated to Python for Larch

import numpy as np
from larch import Make_CallArgs, parse_group_args
from larch.math import complex_phase
from .xafsutils import set_xafsGroup

@Make_CallArgs(["k" ,"chi"])
def cauchy_wavelet(k, chi=None, group=None, kweight=0, rmax_out=10,
                   nfft=2048, _larch=None):
    """
    Cauchy Wavelet Transform for XAFS, following work of Munoz, Argoul, and Farges

    Parameters:
    -----------
      k:        1-d array of photo-electron wavenumber in Ang^-1 or group
      chi:      1-d array of chi
      group:    output Group
      rmax_out: highest R for output data (10 Ang)
      kweight:  exponent for weighting spectra by k**kweight
      nfft:     value to use for N_fft (2048).

      Returns:
    ---------
      None   -- outputs are written to supplied group.

    Notes:
    -------
    Arrays written to output group:
    r                  uniform array of R, out to rmax_out.
    wcauchy            complex cauchy wavelet(k, R)
    wcauchy_mag        magnitude of wavelet(k, R)
    wcauchy_re         real part of wavelet(k, R)
    wcauchy_im         imaginary part of wavelet(k, R)

    Supports First Argument Group convention (with group
    member names 'k' and 'chi')

    """
    k, chi, group = parse_group_args(k, members=('k', 'chi'),
                                     defaults=(chi,), group=group,
                                     fcn_name='cauchy_wavelet')

    kstep = np.round(1000.*(k[1]-k[0]))/1000.0
    rstep = (np.pi/2048)/kstep
    rmin = 1.e-7
    rmax = rmax_out
    nrpts = int(np.round((rmax-rmin)/rstep))
    nkout = len(k)
    if kweight != 0:
        chi = chi * k**kweight

    # extend EXAFS to 1024 data points...
    NFT = int(nfft/2)
    if len(k) < NFT:
        knew = np.arange(NFT) * kstep
        xnew = np.zeros(NFT) * kstep
        xnew[:len(k)] = chi
    else:
        knew = k[:NFT]
        xnew = chi[:NFT]

    # FT parameters
    freq = (1.0/kstep)*np.arange(nfft)/(2*nfft)
    omega = 2*np.pi*freq

    # simple FT calculation
    tff = np.fft.fft(xnew, n= 2*nfft)

    # scale parameter
    r  = np.linspace(0, rmax, nrpts)
    r[0] = 1.e-19
    a  = nrpts/(2*r)

    # Characteristic values for Cauchy wavelet:
    cauchy_sum = np.log(2*np.pi) - np.log(1.0+np.arange(nrpts)).sum()

    # Main calculation:
    out = np.zeros(nkout*nrpts,
                   dtype='complex128').reshape(nrpts, nkout)
    for i in range(nrpts):
        aom = a[i]*omega
        aom[np.where(aom==0)] = 1.e-19
        filt = cauchy_sum + nrpts*np.log(aom) - aom
        tmp  = np.conj(np.exp(filt))*tff[:nfft]
        out[i, :] = np.fft.ifft(tmp, 2*nfft)[:nkout]

    group = set_xafsGroup(group, _larch=_larch)
    group.r  =  r
    group.wcauchy =  out
    group.wcauchy_mag =  np.sqrt(out.real**2 + out.imag**2)
    group.wcauchy_re =  out.real
    group.wcauchy_im =  out.imag



#!/usr/bin/env python
from __future__ import print_function

import time
import sys

class debugtime(object):
    def __init__(self, verbose=False, _larch=None):
        self._larch = _larch
        self.clear()
        self.verbose = verbose
        self.add('init')

    def clear(self):
        self.times = []

    def _print(self, *args):
        writer = sys.stdout
        if self._larch is not None:
            writer = self._larch.writer
        writer.write(*args)


    def add(self,msg=''):
        if self.verbose:
            self._print(msg, time.ctime())
        self.times.append((msg, time.time()))

    def get_report(self):
        m0, t0 = self.times[0]
        tlast= t0
        out = []
        add = out.append
        add("# %s  %s " % (m0,time.ctime(t0)))
        add("#----------------")
        add("#       Message                       Total     Delta")
        for m,t in self.times[1:]:
            tt = t-t0
            dt = t-tlast
            if len(m)<32:
                m = m + ' '*(32-len(m))
            add("  %32s    %.3f    %.3f" % (m,tt, dt))
            tlast = t
        return "\n".join(out)

    def show(self):
        self._print(self.get_report())

    def save(self, fname='debugtimer.dat'):
        dat = self.get_report()
        with open(fname, 'w') as fh:
            fh.write('%s\n' % dat)

def debugtimer(_larch=None):
    """debugtimer returns a Timer object that can be used
    to time the running of portions of code, and then
    write a simple report of the results.  the Timer object
    has methods:

      clear()   -- reset Timer
      add(msg)  -- record time, with message
      show_report -- print timer report

    An example:

      timer = debugtimer()
      x = 1
      timer.add('now run foo')
      foo()
      timer.add('now run bar')
      bar()
      timer.show_report()
    """
    return debugtime(_larch=_larch)



"""
ini doc
"""
__DOC__ = '''
Functions for Input/Output, especially reading specific types
of scientific data files.
'''
from .fileutils import (increment_filename, new_filename, new_dirname,
                        fix_filename, fix_varname, pathOf, unixpath,
                        winpath, nativepath, strip_quotes, get_timestamp,
                        asciikeys)

from .columnfile import (read_ascii, write_ascii, write_group, set_array_labels,
                         guess_filereader, look_for_nans)
from .xdi import read_xdi, XDIFile, XDIFileException
from .mda import read_mda
from .hdf5group import h5file, h5group, netcdf_file, netcdf_group

from .xsp3_hdf5 import read_xsp3_hdf5
from .xrf_netcdf import read_xrf_netcdf
from .xrd_netcdf import read_xrd_netcdf
from .xrd_hdf5 import read_xrd_hdf5

from .xdi import read_xdi

from .gse_escan import gsescan_group, gsescan_deadtime_correct
from .gse_xdiscan import read_gsexdi, gsexdi_deadtime_correct, is_GSEXDI
from .gse_mcafile import gsemca_group, GSEMCA_File
from .save_restore import save, restore
from .tifffile import TIFFfile
from .athena_project import (is_athena_project, read_athena, AthenaProject, create_athena,
                             extract_athenagroup, make_hashkey)

from .xafs_beamlines import guess_beamline

from .csvfiles import groups2csv, read_csv
from .export_modelresult import export_modelresult
from .mergegroups import merge_groups
from .specfile_reader import (HAS_SPECFILE, spec_getscan2group,
                              spec_getmap2group, spec_getmrg2group,
                              str2rng_larch)
from .stepscan_file import read_stepscan
from . import tifffile
def read_tiff(fname, _larch=None, *args, **kws):
    """read image data from a TIFF file as an array"""
    return tifffile.imread(fname, *args, **kws)


__exports__ = dict(increment_filename=increment_filename,
                   new_filename=new_filename,
                   new_dirname=new_dirname,
                   fix_filename=fix_filename,
                   fix_varname=fix_varname,
                   pathOf=pathOf,
                   unixpath=unixpath,
                   winpath=winpath,
                   nativepath=nativepath,
                   strip_quotes=strip_quotes,
                   get_timestamp=get_timestamp,
                   asciikeys=asciikeys,
                   read_ascii=read_ascii,
                   look_for_nans=look_for_nans,
                   set_array_labels=set_array_labels,
                   guess_filereader=guess_filereader,
                   write_ascii=write_ascii,
                   write_group=write_group,
                   groups2csv=groups2csv,
                   read_csv=read_csv,
                   read_xdi=read_xdi,
                   read_athena=read_athena,
                   create_athena=create_athena,
                   extract_athenagroup=extract_athenagroup,
                   export_modelresult=export_modelresult,
                   read_gsescan=gsescan_group,
                   gsescan_dtcorrect=gsescan_deadtime_correct,
                   read_gsemca=gsemca_group,
                   read_gsexdi=read_gsexdi,
                   gsexdi_deadtime_correct=gsexdi_deadtime_correct,
                   read_mda=read_mda,
                   read_stepscan=read_stepscan,
                   read_tiff=read_tiff,
                   merge_groups=merge_groups,
                   save=save,
                   restore=restore,
                   read_xrd_hdf5=read_xrd_hdf5,
                   read_xrd_netcdf=read_xrd_netcdf,
                   read_xrf_netcdf=read_xrf_netcdf,
                   read_xsp3_hdf5=read_xsp3_hdf5,
                   h5group=h5group,
                   h5file=h5file,
                   netcdf_file=netcdf_file,
                   netcdf_group=netcdf_group,
                   )




if HAS_SPECFILE:
    __exports__.update(dict(read_specfile_scan=spec_getscan2group,
                            read_specfile_map=spec_getmap2group,
                            read_specfile_mrg=spec_getmrg2group,
                            str2rng=str2rng_larch))

_larch_builtins = {'_io':__exports__}
